##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for ControlVolumeBlockData.

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value)

from idaes.core import FlowsheetBlock, EnergyBalanceType, MomentumBalanceType
from idaes.unit_models.gibbs_reactor import GibbsReactor
#from idaes.property_models.examples.methane_combustion_ideal import (
#    MethaneCombustionParameterBlock)
from idaes.property_models.activity_coeff_models.methane_combustion_ideal \
    import MethaneParameterBlock as MethaneCombustionParameterBlock
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = GibbsReactor(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 8

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def methane(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = MethaneCombustionParameterBlock()

        m.fs.unit = GibbsReactor(default={
                "property_package": m.fs.properties,
                "has_heat_transfer": True,
                "has_pressure_change": True})

        return m

    @pytest.mark.build
    def test_build(self, methane):
        assert hasattr(methane.fs.unit, "inlet")
        assert len(methane.fs.unit.inlet.vars) == 4
        assert hasattr(methane.fs.unit.inlet, "flow_mol")
        assert hasattr(methane.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.inlet, "temperature")
        assert hasattr(methane.fs.unit.inlet, "pressure")

        assert hasattr(methane.fs.unit, "outlet")
        assert len(methane.fs.unit.outlet.vars) == 4
        assert hasattr(methane.fs.unit.outlet, "flow_mol")
        assert hasattr(methane.fs.unit.outlet, "mole_frac_comp")
        assert hasattr(methane.fs.unit.outlet, "temperature")
        assert hasattr(methane.fs.unit.outlet, "pressure")

        assert hasattr(methane.fs.unit, "gibbs_minimization")
        assert hasattr(methane.fs.unit, "heat_duty")
        assert hasattr(methane.fs.unit, "deltaP")

        assert number_variables(methane) == 80
        assert number_total_constraints(methane) == 67
        assert number_unused_variables(methane) == 0

    def test_dof(self, methane):
        methane.fs.unit.inlet.flow_mol[0].fix(230.0)
        methane.fs.unit.inlet.mole_frac_comp[0, "H2"].fix(0.0435)
        methane.fs.unit.inlet.mole_frac_comp[0, "N2"].fix(0.6522)
        methane.fs.unit.inlet.mole_frac_comp[0, "O2"].fix(0.1739)
        methane.fs.unit.inlet.mole_frac_comp[0, "CO2"].fix(1e-5)
        methane.fs.unit.inlet.mole_frac_comp[0, "CH4"].fix(0.1304)
        methane.fs.unit.inlet.mole_frac_comp[0, "CO"].fix(1e-5)
        methane.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix(1e-5)
        methane.fs.unit.inlet.mole_frac_comp[0, "NH3"].fix(1e-5)
        methane.fs.unit.inlet.temperature[0].fix(1500.0)
        methane.fs.unit.inlet.pressure[0].fix(101325.0)

        methane.fs.unit.outlet.temperature[0].fix(2844.38)
        methane.fs.unit.deltaP.fix(0)

        assert degrees_of_freedom(methane) == 0

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize_temperature(self, methane):
        orig_fixed_vars = fixed_variables_set(methane)
        orig_act_consts = activated_constraints_set(methane)

        methane.fs.unit.initialize(
                          optarg={'tol': 1e-6},
                          state_args={'temperature': 2844.38,
                                      'pressure': 101325.0,
                                      'flow_mol': 251.05,
                                      'mole_frac_comp': {'CH4': 1e-5,
                                                         'CO': 0.0916,
                                                         'CO2': 0.0281,
                                                         'H2': 0.1155,
                                                         'H2O': 0.1633,
                                                         'N2': 0.5975,
                                                         'NH3': 1e-5,
                                                         'O2': 0.0067}})

        assert degrees_of_freedom(methane) == 0

        fin_fixed_vars = fixed_variables_set(methane)
        fin_act_consts = activated_constraints_set(methane)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve_temperature(self, methane):
        results = solver.solve(methane)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution_temperature(self, methane):
        assert (pytest.approx(250.06, abs=1e-2) ==
                value(methane.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(0.0, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]))
        assert (pytest.approx(0.0974, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO"]))
        assert (pytest.approx(0.0226, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]))
        assert (pytest.approx(0.1030, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2"]))
        assert (pytest.approx(0.1769, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]))
        assert (pytest.approx(0.5999, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "N2"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]))
        assert (pytest.approx(0.0002, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "O2"]))
        assert (pytest.approx(-7454077, abs=1e2) ==
                value(methane.fs.unit.heat_duty[0]))
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(methane.fs.unit.outlet.pressure[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation_temperature(self, methane):
        assert abs(value(
                methane.fs.unit.inlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase["Vap"] -
                methane.fs.unit.outlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase["Vap"] +
                methane.fs.unit.heat_duty[0])) <= 1e-6

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize_duty(self, methane):
        methane.fs.unit.outlet.temperature[0].unfix()
        methane.fs.unit.heat_duty.fix(-7454077)
        assert degrees_of_freedom(methane) == 0

        orig_fixed_vars = fixed_variables_set(methane)
        orig_act_consts = activated_constraints_set(methane)

        methane.fs.unit.initialize(
                          optarg={'tol': 1e-6},
                          state_args={'temperature': 2844.38,
                                      'pressure': 101325.0,
                                      'flow_mol': 251.05,
                                      'mole_frac_comp': {'CH4': 1e-5,
                                                         'CO': 0.0916,
                                                         'CO2': 0.0281,
                                                         'H2': 0.1155,
                                                         'H2O': 0.1633,
                                                         'N2': 0.5975,
                                                         'NH3': 1e-5,
                                                         'O2': 0.0067}})

        assert degrees_of_freedom(methane) == 0

        fin_fixed_vars = fixed_variables_set(methane)
        fin_act_consts = activated_constraints_set(methane)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve_heat_duty(self, methane):
        results = solver.solve(methane)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution_duty(self, methane):
        assert (pytest.approx(250.06, abs=1e-2) ==
                value(methane.fs.unit.outlet.flow_mol[0]))
        assert (pytest.approx(0.0, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CH4"]))
        assert (pytest.approx(0.0974, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO"]))
        assert (pytest.approx(0.0226, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "CO2"]))
        assert (pytest.approx(0.1030, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2"]))
        assert (pytest.approx(0.1769, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "H2O"]))
        assert (pytest.approx(0.5999, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "N2"]))
        assert (pytest.approx(0.0, abs=1e-5) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "NH3"]))
        assert (pytest.approx(0.0002, abs=1e-4) ==
                value(methane.fs.unit.outlet.mole_frac_comp[0, "O2"]))
        assert (pytest.approx(-7454077, abs=1e2) ==
                value(methane.fs.unit.heat_duty[0]))
        assert (pytest.approx(101325.0, abs=1e-2) ==
                value(methane.fs.unit.outlet.pressure[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation_duty(self, methane):
        assert abs(value(
                methane.fs.unit.inlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_in[0]
                    .enth_mol_phase["Vap"] -
                methane.fs.unit.outlet.flow_mol[0] *
                methane.fs.unit.control_volume.properties_out[0]
                    .enth_mol_phase["Vap"] +
                methane.fs.unit.heat_duty[0])) <= 1e-6

    @pytest.mark.ui
    def test_report(self, methane):
        methane.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Index Property metadata
"""
# stdlib
import logging
# local
import idaes
from idaes.dmf import codesearch
from idaes.dmf import resource
from idaes.core import property_base  # noqa: F401

__author__ = 'Dan Gunter <dkgunter@lbl.gov>'

_log = logging.getLogger(__name__)


def index_property_metadata(dmf, pkg=idaes, expr='_PropertyMetadata.*',
                            default_version='0.0.1', **kwargs):
    """Index all the PropertyMetadata classes in this package.

    Usually the defaults will be correct, but you can modify the package
    explored and set of classes indexed.

    When you re-index the same class (in the same module), whether or
    not that is a "duplicate" will depend on the version found in the
    containing module. If there is no version in the containing module,
    the default version is used (so it is always the same). If it is
    a duplicate, nothing is done, this is not considered an error. If a new
    version is added, it will be explicitly connected to the highest version
    of the same module/code. So, for example,

    1. Starting with (a.module.ClassName version=0.1.2)
    2. If you then find a new version (a.module.ClassName version=1.2.3)
       There will be 2 resources, and you will have the relation::

           a.module.ClassName/1.2.3 --version---> a.module.ClassName/0.1.2

    3. If you add another version (a.module.ClassName version=1.2.4), you
       will have two relations::

           a.module.ClassName/1.2.3 --version---> a.module.ClassName/0.1.2
           a.module.ClassName/1.2.4 --version---> a.module.ClassName/1.2.3

    Args:
        dmf (idaes.dmf.DMF): Data Management Framework instance in which
                             to record the found metadata.
        pkg (module): Root module (i.e. package root) from which to find
                      the classes containing metadata.
        expr (str): Regular expression pattern for the names of the classes
                    in which to look for metadata.
        default_version (str): Default version to use for modules
                    with no explicit version.
        kwargs: Other keyword arguments passed to
                      :class:`codesearch.ModuleClassWalker`.
    Returns:
        codesearch.ModuleClassWalker: Class that walked through the modules.
            You can call `.get_indexed_classes()` to see the list of classes
            walked, or `.walk()` to walk the modules again.
    Raises:
        This instantiated a `DMFVisitor` and calls its `walk()` method to
        walk/visit each found class, so any exception raised by the constructor
        or `DMFVisitor.visit_metadata()`.
    """
    wlk = codesearch.ModuleClassWalker(
        from_pkg=pkg, class_expr=expr,
        parent_class=idaes.core.property_meta.HasPropertyClassMetadata,
        suppress_warnings=True, **kwargs)
    vst = DMFVisitor(dmf, default_version=default_version)
    wlk.walk(vst)
    return wlk


class DMFVisitor(codesearch.PropertyMetadataVisitor):

    #: Added to resource 'tags', so easier to find later
    INDEXED_PROPERTY_TAG = 'indexed-property'

    def __init__(self, dmf, default_version=None):
        """Constructor.

        Args:
            dmf (idaes.dmf.DMF): Data management framework.
            default_version (Union[None, list, tuple]): Default version to give
                 the class, if the containing module does not have
                 a `__version__` variable. If None, the absence of that
                 variable will cause an error.
        Raises:
            TypeError: if `default_version` isn't something that
                       :func:`resource.version_list` can convert.
        """
        self._dmf = dmf
        self._defver = None
        if default_version is not None:
            try:
                self._defver = resource.version_list(default_version)
            except ValueError as err:
                raise TypeError('Bad "default_version": {}'.format(err))

    def visit_metadata(self, obj, meta):
        """Called for each property class encountered during the "walk"
         initiated by `index_property_metadata()`.

        Args:
            obj (property_base.PropertyParameterBase): Property class instance
            meta (property_base.PropertyClassMetadata): Associated metadata
        Returns:
            None
        Raises:
            AttributeError: if
        """
        _log.debug('Adding resource to DMF that indexes the property package '
                   '"{}"'.format('.'.join([obj.__module__, obj.__name__])))
        r = resource.Resource()
        r.v[r.TYPE_FIELD] = resource.TY_CODE
        r.data = {'units': meta.default_units,
                  'properties': meta.properties}
        containing_module = obj.__module__
        if hasattr(containing_module, '__version__'):
            obj_ver = resource.version_list(containing_module.__version__)
        elif self._defver is None:
            raise AttributeError('No __version__ for module {}, and no '
                                 'default'.format(containing_module))
        else:
            obj_ver = self._defver
        r.v['codes'].append({
            'type': 'class',
            'language': 'python',
            'name': '.'.join([obj.__module__, obj.__name__]),
            'version': obj_ver
        })
        r.v['tags'].append(self.INDEXED_PROPERTY_TAG)
        # Search for existing indexed codes.
        # A match exists if all 3 of these are the same:
        #   codes.type == class
        #   codes.language == python
        #   codes.name == <module>.<class>
        info = {k: r.v['codes'][0][k] for k in ('type', 'language', 'name')}
        rsrc_list, dup_rsrc = [], None
        # Loop through all the right kind of resources
        for rsrc in self._dmf.find({r.TYPE_FIELD: resource.TY_CODE,
                                    'tags': ['indexed-property']}):
            # skip any resources without one code
            if len(rsrc.v['codes']) != 1:
                continue
            code = rsrc.v['codes'][0]
            # skip any resource of wrong code type, name, lang.
            skip = False
            for k in info:
                if code[k] != info[k]:
                    skip = True
                    break
            if skip:
                continue
            # skip any resources missing the recorded metadata
            skip = False
            for data_key in r.data.keys():
                if data_key not in rsrc.data:
                    skip = True
                    break
            if skip:
                continue
            # If the version of the found code is the same as the
            # version of the one to be added, then it is a duplicate
            if code['version'] == obj_ver:
                dup_rsrc = rsrc
                break
            rsrc_list.append(rsrc)
        if dup_rsrc:
            # This is considered a normal, non-exceptional situation
            _log.debug('DMFVisitor: Not adding duplicate index for '
                       '{}v{}'.format(info['name'], obj_ver))
        else:
            # add the resource
            r.validate()
            _log.debug('DMFVisitor: Adding resource for code "{}"v{} type={}'
                       .format(r.v['codes'][0]['name'],
                               r.v['codes'][0]['version'],
                               r.v['codes'][0]['type']))

            self._dmf.add(r)
            if rsrc_list:
                # Connect to most recent (highest) version
                rsrc_list.sort(key=lambda rs: rs.v['codes'][0]['version'])
                # for rsrc in rsrc_list:
                rsrc = rsrc_list[-1]
                rel = resource.Triple(r, resource.PR_VERSION, rsrc)
                resource.create_relation(rel)
                self._dmf.update(rsrc)
                self._dmf.update(r)



# -*- coding: UTF-8 -*-
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This module contains utility functions useful for validating arguments to
IDAES modeling classes. These functions are primarily designed to be used as
the `domain` argument in ConfigBlocks.
"""

__author__ = "Andrew Lee"

from pyomo.environ import Set
from pyomo.dae import ContinuousSet
from pyomo.network import Port
from idaes.core import useDefault
from idaes.core.util.exceptions import ConfigurationError
import logging

_log = logging.getLogger(__name__)


def is_physical_parameter_block(val):
    '''Domain validator for property package attributes

    Args:
        val : value to be checked

    Returns:
        ConfigurationError if val is not an instance of PhysicalParameterBlock
        or useDefault
    '''
    from idaes.core.property_base import PhysicalParameterBlock
    if isinstance(val, PhysicalParameterBlock) or val == useDefault:
        return val
    else:
        _log.error("Property package argument {} should == useDefault or "
                   "be an instance of PhysicalParameterBlock".format(val))
        raise ConfigurationError(
                """Property package argument should be an instance
                of a PhysicalParameterBlock or useDefault""")


def is_reaction_parameter_block(val):
    '''Domain validator for reaction package attributes

    Args:
        val : value to be checked

    Returns:
        ConfigurationError if val is not an instance of ReactionParameterBlock
    '''
    from idaes.core.reaction_base import ReactionParameterBlock
    if isinstance(val, ReactionParameterBlock):
        return val
    else:
        raise ConfigurationError(
                """Reaction package argument should be an instance
                of a ReactionParameterBlock""")


def is_state_block(val):
    '''Domain validator for state block as an argument

    Args:
        val : value to be checked

    Returns:
        ConfigurationError if val is not an instance of StateBlock
        or None
    '''
    from idaes.core.property_base import StateBlock
    if (isinstance(val, StateBlock) or val is None):
        return val
    else:
        raise ConfigurationError(
                """State block should be an instance of a StateBlock or
                None""")


def list_of_floats(arg):
    '''Domain validator for lists of floats

    Args:
        arg : argument to be cast to list of floats and validated

    Returns:
        List of strings
    '''
    try:
        # Assume arg is iterable
        lst = [float(i) for i in arg]
    except TypeError:
        # arg is not iterable
        lst = [float(arg)]
    return lst


def list_of_strings(arg):
    '''Domain validator for lists of strings

    Args:
        arg : argument to be cast to list of strings and validated

    Returns:
        List of strings
    '''
    if isinstance(arg, dict):
        raise ConfigurationError("Invalid argument type (dict). "
                                 "Expected a list of strings, or something "
                                 "that can be cast to a list of strings")

    try:
        # Assume arg is iterable
        if isinstance(arg, str):
            lst = [arg]
        else:
            lst = [str(i) for i in arg]
    except TypeError:
        # arg is not iterable
        lst = [str(arg)]
    return lst


def is_port(arg):
    '''Domain validator for ports

    Args:
        arg : argument to be checked as a Port

    Returns:
        Port object or Exception
    '''
    if not isinstance(arg, Port):
        raise ConfigurationError('Invalid argument type. Expected an instance '
                                 'of a Pyomo Port object')
    return arg


def is_time_domain(arg):
    '''Domain validator for time domains

    Args:
        arg : argument to be checked as a time domain (i.e. Set or
        ContinuousSet)

    Returns:
        Set, ContinuousSet or Exception
    '''
    if not isinstance(arg, (Set, ContinuousSet)):
        raise ConfigurationError('Invalid argument type. Expected an instance '
                                 'of a Pyomo Set or ContinuousSet object')
    return arg


def is_transformation_method(arg):
    '''Domain validator for transformation methods

    Args:
        arg : argument to be checked for membership in recognized strings

    Returns:
        Recognised string or Exception
    '''
    if arg in ["dae.finite_difference",
               "dae.collocation"]:
        return arg
    else:
        raise ConfigurationError(
                'Invalid value provided for transformation_method. '
                'Please check the value and spelling of the argument provided.'
                )


def is_transformation_scheme(arg):
    '''Domain validator for transformation scheme

    Args:
        arg : argument to be checked for membership in recognized strings

    Returns:
        Recognised string or Exception
    '''
    if arg in ["BACKWARD", "FORWARD", "LAGRANGE-RADAU", "LAGRANGE-LEGENDRE"]:
        return arg
    else:
        raise ConfigurationError(
                'Invalid value provided for transformation_scheme. '
                'Please check the value and spelling of the argument provided.'
                )



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def almpickle(data, vargs):
    import pickle
    import sys
    # remove lambda function from results dictionary
    if (len(vargs) == 1):
        fname = vargs[0]
    elif (len(vargs) == 2):
        fname = vargs[0]
        dopick = vargs[1]
    else:
        fname = 'alamo.pick'
        dopick = True

    if (not isinstance(data, list)):
        if ('f(model)' in data.keys()):
            data['f(model)'] = ()
        else:
            sys.stdout.write('Results dictionary does not contain lambda function')

    if (dopick):
        pickle.dump(data, open(fname, "wb"))


def postpickle(data):
    # Recompile lambda function from model string and labels
    # import sympy
    from sympy.parsing.sympy_parser import parse_expr
    from sympy import symbols, lambdify

    if (isinstance(data['model'], type({}))):
        data['f(model)'] = {}
        for olab in data['zlabels']:
            model_str = data['model'].split('=')[1].replace('^', '**')
            data['f(model)'][olab] = lambdify([symbols(data['xlabels'])], 
                                              parse_expr(model_str), "numpy")
    else:
        model_str = data['model'].split('=')[1].replace('^', '**')
        data['f(model)'] = lambdify([symbols(data['xlabels'])],
                                    parse_expr(model_str), "numpy")

    return data


def almunpickle(fname):
    # unpickle and relambdify
    from idaes.surrogate import alamopy
    # from alamopy import postpickle
    import pickle

    res = pickle.load(open(fname, 'rb'))
    if ('f(model)' in res[0].keys()):
        res = alamopy.postpickle(res)
    return res



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np
from . import mechs
from .shared import sharedata as sd
# from ripe.shared import debug as debug

consivars = sd["ivars"]


def makeaterm(data, stoich, mechs, kwargs, ncons, mechlist, fixarray, sharedata):
    # This subroutine constructs an activity matrix that is used to encode
    # considered stoichimetry and mechanisms
    # Input :
    # data      - process data from ripemodel()
    # stoichs   - considered reactions toichiometries
    # mechs     - considered reaction mechanisms
    # kwargs    - keyword arguments from ripemodel()
    # mechlist  - list containing which mechs correspond to which stoich
    # fixarray  - [#mech by #stoich] matrix that encodes valid combinations
    # Outputs :
    # aterm     - activity matrix
    # fdata     - formatted data for use in writing pyomo models
    # pc        - formatted process conditions
    # data      - data is reshaped and returned
    # scales    -  dictionary containing relevant information for scaling

    # Call subroutine to analyze kwargs and format data
    data, kwargs, fdata, pc, alld = formatinputs(data, kwargs)

    # Find sizes and initialize aterm
    ndata, ns = np.shape(fdata)
    nm = len(mechlist)
    nh = len(stoich)
    aterm = np.zeros([ndata, ns, nm, nh])

    # This section of code loops thorugh the aterm array
    # Callling s_mech in order to determine numerical values
    # mechanisms can be specified with variable stoichiometry
    for i1 in range(ndata):
        # index i1 over observations
        for i2 in range(ns):
            # index i2 over species
            i3 = 0
            # index i3 over mechanisms
            for tempi in range(len(mechs)):
                mechline = mechs[tempi]
                for mspec in mechline[1]:
                    for h_ind in list(mechline[0]):
                        # final index over stoichiometries
                        if mspec == "massact" or mechline[2]:
                            s_mech = mechs.mechperstoich(mspec, stoich[h_ind])
                        else:
                            s_mech = mspec
                        if "T" in kwargs.keys():
                            inv = np.hstack((fdata[i1, :], pc["T"][i1][0]))
                        else:
                            inv = fdata[i1, :]
                        atemp = stoich[h_ind][i2] * s_mech(*inv)
                        aterm[i1, i2, i3, h_ind] = atemp
                    i3 += 1

    # Scale data if specified
    if sharedata["ascale"]:
        aterm, scales = normalizefeatures(aterm, fixarray)
    else:
        scales = []

    return [aterm, fdata, pc, data, scales]


def formatinputs(data, kwargs):
    # This subroutine formats inputs supplied to ripemodel()
    # Inputs:
    # data     - Input data that contains variable size and type
    # kwargs   - key word arguments supplied to ripemodel()
    # Outputs:
    # data     - Format and shape of input data is made consistent
    # kwargs   - kwargs is checked and returned
    # fdata    - formatted input data
    # pc       - process conditions are puleld from kwargs and formatted
    # alldata  - another reformatting of existing data required by ripeems()

    # input data may be 1,2, or 3 dimensional at this point

    # Define what kwargs are used
    inkeys = list(kwargs)  # kwargs.keys() python 2
    # Define what keys are not used, dynamic data is handled differently so t is removed
    inkeys_not = list(set(inkeys) - set(["t", "other"]))
    # This line ensures additional kwargs are not analyzed/can be defined
    notinkeys = list(set(consivars) - set(inkeys))

    # initialize process condition dictionary and find shape of data provided
    pc = {}
    dshape = np.shape(data)

    if "t" in inkeys:
        # dictionaries do not perserve order - ensure 't' is at end
        inkeys.remove("t")
        inkeys.append("t")
        # if data is 1 dimensional and dynamic then no process conditions should be specified
        if len(dshape) == 1:
            data = np.expand_dims(np.expand_dims(data, axis=-1), axis=-1)
            ns = 1
            nobs = dshape[0]
            npc = 1
        elif len(dshape) == 2:
            data = np.expand_dims(data, axis=-1)
            # first test to see if other process conditions exist
            if inkeys == inkeys_not:
                # ripe problem is dynamic with 1 process condition
                ns = dshape[1]
                nobs = dshape[0]
                npc = 1
            else:
                # data must have only 1 species
                ns = 1
                npc = dshape[1]
                nobs = dshape[0]
        else:
            nobs = dshape[0]
            ns = dshape[1]
            npc = dshape[2]
    else:
        # nobs is used to track observations per process condition without time this is 1
        nobs = 1
        if len(dshape) == 1:
            # data has one species
            data = np.expand_dims(np.expand_dims(data, axis=-1), axis=-1)
            npc = dshape[0]
            ns = 1
        else:
            # third dimension is redudant be needed for future code
            data = np.expand_dims(np.expand_dims(data, axis=-1), axis=-1)
            ns = dshape[1]
            npc = dshape[0]

    # ensure consistency in process condition formats
    kwargs = checkargs(kwargs, inkeys, npc, nobs, ns)

    # ndata is always npc*nobs
    ndata = npc * nobs

    # initialize formatted data array
    fdata = np.zeros([ndata, ns])

    # Construct process condition array
    for key in inkeys_not:
        pc[key] = kwargs[key]

    # construction of fdata makes construction of aterm much easier
    ind = 0
    if "t" in inkeys:
        pc["t"] = np.ndarray.flatten(kwargs["t"])
        if npc == 1 and ns == 1:
            for i in range(nobs):
                fdata[i, :] = data[i, 0, 0]
        elif npc == 1:
            for i in range(nobs):
                for j in range(ns):
                    fdata[i, j] = data[i, j, 0]
        elif ns == 1:
            while ind < ndata:
                for i in range(nobs):
                    for j in range(npc):
                        fdata[ind, 0] = data[i, j, 0]
                        ind += 1
        else:
            while ind < ndata:
                for k in range(npc):
                    for i in range(nobs):
                        for j in range(ns):
                            fdata[ind, j] = data[i, j, k]
                        ind += 1
    else:
        if ns == 1:
            for i in range(npc):
                fdata[i, 0] = data[i, 0, 0]
        else:
            for i in range(npc):
                for j in range(ns):
                    fdata[i, j] = data[i, j, 0]

    # Initialize notinkeys for easy code in pyomo models
    for key in notinkeys:
        if key == "vol":
            pc["vol"] = [1.0] * ndata
        elif key == "flow":
            pc["flow"] = [[1.0] * ns] * ndata
        elif key == "x0":
            pc["x0"] = [[0.0] * ns] * ndata
        elif key == "T":
            pc["T"] = [[-1.0]] * ndata

    # define npc and nobs for easy access later
    #  additional process-specific information could
    #  be define in the dictionary pc without issue
    pc["npc"] = npc
    pc["nobs"] = nobs

    if "Tr" in kwargs.keys():
        pc["Tref"] = kwargs["Tr"][0][0]

    # return concatenated array for use in returned models in ems
    num_other = 0
    alldata = np.zeros([ndata, 3 * ns + 2 + num_other])
    for i in range(ndata):
        alldata[i, :ns] = fdata[i, :]
        alldata[i, ns : 2 * ns] = pc["x0"][i]
        alldata[i, 2 * ns] = pc["T"][i][0]
        alldata[i, 2 * ns + 1 : 3 * ns + 1] = pc["flow"][:][i]
        alldata[i, 3 * ns + 1] = pc["vol"][i]

    # append other process conditions
    if "other" in pc.keys():
        alldata = np.concatenate((alldata, pc["other"]), axis=0)

    return [data, kwargs, fdata, pc, alldata]


def checkargs(kwargs, inkeys, npc, nobs, ns):
    # This subroutine checks and formats kwargs in order to construct a process condition array
    # Inputs:
    # kwargs  - kwargs from ripemodel()
    # inkeys  - kwargs specified
    # npc     - number of process conditions
    # nobs    - number of observations
    # ns      - number of species
    # Outputs:
    # kwargs - reformatted kwargs

    for key in list(set(inkeys) - set(["t", "Tref"])):
        # reformat kwargs so it is easier to process
        if isinstance(kwargs[key], type(0.0)):
            kwargs[key] = [[kwargs[key]]] * npc
        elif isinstance(kwargs[key], type(0)):
            kwargs[key] = [[float(kwargs[key])]] * nobs
        elif key in ["x0", "flow"]:
            if len(np.shape(kwargs[key])) == 1 and ns != 1:
                kwargs[key] = [kwargs[key]] * npc
            elif len(np.shape(kwargs[key])) == 1 and ns == 1:
                kwargs[key] = np.expand_dims(np.asarray(kwargs[key]), axis=-1)
        else:
            kwargs[key] = np.expand_dims(np.asarray(kwargs[key]), axis=-1)
    if "t" in kwargs:
        if len(np.shape(kwargs["t"])) == 1 and npc != 1:
            kwargs["t"] = [kwargs["t"]] * npc

    return kwargs


def normalizefeatures(aterm, fixarray):
    # This subroutine calculate scaling values for an activity array
    # Inputs:
    # aterm      - activity array
    # fixarray   - array containing mech/stoich pairs
    # Outputs:
    # aterm      - scaled aterm
    # scale_dict - information for unwinding scales

    nd, ns, nm, nh = np.shape(aterm)

    scale_dict = {}
    scale_dict["aterm_u"] = aterm
    n_factor = np.zeros([nm, nh])
    for k in range(nm):
        for l in range(nh):  # L
            if fixarray[l, k] == 1:
                maxval = np.max(
                    [x for x in np.ndarray.flatten(aterm[:, :, k, l])]
                )  # if x > 0.0])
                minval = np.min(
                    [x for x in np.ndarray.flatten(aterm[:, :, k, l])]
                )  # if x > 0.0])
                n_factor[k, l] = np.max([np.abs(maxval), np.abs(minval)])
                for i in range(nd):
                    for j in range(ns):
                        aterm[i, j, k, l] = aterm[i, j, k, l] / n_factor[k, l]
    scale_dict["nfactor"] = n_factor
    return aterm, scale_dict



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Flash unit model.
Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           value)

from idaes.core import (FlowsheetBlock, MaterialBalanceType, EnergyBalanceType,
                        MomentumBalanceType)
from idaes.unit_models.flash import Flash, EnergySplittingType
from idaes.property_models.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.property_models import iapws95
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_unused_variables)
from idaes.core.util.testing import (get_default_solver,
                                     PhysicalParameterTestBlock)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


# -----------------------------------------------------------------------------
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = Flash(default={"property_package": m.fs.properties})

    # Check unit config arguments
    assert len(m.fs.unit.config) == 11

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == \
        MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == \
        EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.ideal_separation
    assert m.fs.unit.config.has_heat_transfer
    assert m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.energy_split_basis == \
        EnergySplittingType.equal_temperature


# -----------------------------------------------------------------------------
class TestBTXIdeal(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = BTXParameterBlock(default={"valid_phase":
                                                     ('Liq', 'Vap'),
                                                     "activity_coeff_model":
                                                     "Ideal"})

        m.fs.unit = Flash(default={"property_package": m.fs.properties})

        return m

    @pytest.mark.build
    def test_build(self, btx):
        assert hasattr(btx.fs.unit.inlet, "flow_mol")
        assert hasattr(btx.fs.unit.inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.inlet, "temperature")
        assert hasattr(btx.fs.unit.inlet, "pressure")

        assert hasattr(btx.fs.unit, "vap_outlet")
        assert len(btx.fs.unit.vap_outlet.vars) == 4
        assert hasattr(btx.fs.unit.vap_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.vap_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.vap_outlet, "temperature")
        assert hasattr(btx.fs.unit.vap_outlet, "pressure")

        assert hasattr(btx.fs.unit, "liq_outlet")
        assert len(btx.fs.unit.liq_outlet.vars) == 4
        assert hasattr(btx.fs.unit.liq_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.liq_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.liq_outlet, "temperature")
        assert hasattr(btx.fs.unit.liq_outlet, "pressure")

        assert hasattr(btx.fs.unit, "split")
        assert hasattr(btx.fs.unit, "heat_duty")
        assert hasattr(btx.fs.unit, "deltaP")

        assert number_variables(btx) == 48
        assert number_total_constraints(btx) == 41
        assert number_unused_variables(btx) == 0

    def test_dof(self, btx):
        btx.fs.unit.inlet.flow_mol.fix(1)
        btx.fs.unit.inlet.temperature.fix(368)
        btx.fs.unit.inlet.pressure.fix(101325)
        btx.fs.unit.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        btx.fs.unit.inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        btx.fs.unit.heat_duty.fix(0)
        btx.fs.unit.deltaP.fix(0)

        assert degrees_of_freedom(btx) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, btx):
        orig_fixed_vars = fixed_variables_set(btx)
        orig_act_consts = activated_constraints_set(btx)

        btx.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(btx) == 0

        fin_fixed_vars = fixed_variables_set(btx)
        fin_act_consts = activated_constraints_set(btx)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, btx):
        assert (pytest.approx(0.603, abs=1e-3) ==
                value(btx.fs.unit.liq_outlet.flow_mol[0]))
        assert (pytest.approx(0.396, abs=1e-3) ==
                value(btx.fs.unit.vap_outlet.flow_mol[0]))
        assert (pytest.approx(368, abs=1e-3) ==
                value(btx.fs.unit.liq_outlet.temperature[0]))
        assert (pytest.approx(101325, abs=1e-3) ==
                value(btx.fs.unit.liq_outlet.pressure[0]))

        assert (pytest.approx(0.412, abs=1e-3) ==
                value(btx.fs.unit.liq_outlet.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.588, abs=1e-3) ==
                value(btx.fs.unit.liq_outlet.mole_frac_comp[0, "toluene"]))
        assert (pytest.approx(0.634, abs=1e-3) ==
                value(btx.fs.unit.vap_outlet.mole_frac_comp[0, "benzene"]))
        assert (pytest.approx(0.366, abs=1e-3) ==
                value(btx.fs.unit.vap_outlet.mole_frac_comp[0, "toluene"]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, btx):
        assert abs(value(btx.fs.unit.inlet.flow_mol[0] -
                         (btx.fs.unit.vap_outlet.flow_mol[0] +
                          btx.fs.unit.liq_outlet.flow_mol[0]))) <= 1e-6

    @pytest.mark.ui
    def test_report(self, btx):
        btx.fs.unit.report()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
class TestIAPWS(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.properties = iapws95.Iapws95ParameterBlock(default={
                "phase_presentation": iapws95.PhaseType.LG})

        m.fs.unit = Flash(default={
                "property_package": m.fs.properties,
                "ideal_separation": False,
                "energy_split_basis": EnergySplittingType.enthalpy_split})

        return m

    @pytest.mark.build
    def test_build(self, iapws):
        assert len(iapws.fs.unit.inlet.vars) == 3
        assert hasattr(iapws.fs.unit.inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.inlet, "pressure")

        assert hasattr(iapws.fs.unit, "vap_outlet")
        assert len(iapws.fs.unit.vap_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.vap_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.vap_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.vap_outlet, "pressure")

        assert hasattr(iapws.fs.unit, "liq_outlet")
        assert len(iapws.fs.unit.liq_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.liq_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.liq_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.liq_outlet, "pressure")

        assert hasattr(iapws.fs.unit, "split")
        assert hasattr(iapws.fs.unit, "heat_duty")
        assert hasattr(iapws.fs.unit, "deltaP")

        assert number_variables(iapws) == 18
        assert number_total_constraints(iapws) == 13
        assert number_unused_variables(iapws) == 0

    def test_dof(self, iapws):
        iapws.fs.unit.inlet.flow_mol.fix(100)
        iapws.fs.unit.inlet.enth_mol.fix(24000)
        iapws.fs.unit.inlet.pressure.fix(101325)

        iapws.fs.unit.heat_duty.fix(0)
        iapws.fs.unit.deltaP.fix(0)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.initialization
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_initialize(self, iapws):
        orig_fixed_vars = fixed_variables_set(iapws)
        orig_act_consts = activated_constraints_set(iapws)

        iapws.fs.unit.initialize(optarg={'tol': 1e-6})

        assert degrees_of_freedom(iapws) == 0

        fin_fixed_vars = fixed_variables_set(iapws)
        fin_act_consts = activated_constraints_set(iapws)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_solution(self, iapws):
        t_in = value(iapws.fs.unit.control_volume.properties_in[0].temperature)
        assert (pytest.approx(373.12, abs=1e-2) == t_in)
        assert (pytest.approx(t_in, abs=1e-2) == value(
                iapws.fs.unit.control_volume.properties_out[0].temperature))
        assert (pytest.approx(t_in, abs=1e-2) == value(
                iapws.fs.unit.split.Vap_state[0].temperature))
        assert (pytest.approx(t_in, abs=1e-2) == value(
                iapws.fs.unit.split.Liq_state[0].temperature))

        assert (pytest.approx(101325.0, abs=1e3) ==
                value(iapws.fs.unit.vap_outlet.pressure[0]))
        assert (pytest.approx(48200, abs=1e3) ==
                value(iapws.fs.unit.vap_outlet.enth_mol[0]))
        assert (pytest.approx(40.467, abs=1e-3) ==
                value(iapws.fs.unit.vap_outlet.flow_mol[0]))

        assert (pytest.approx(101325.0, abs=1e3) ==
                value(iapws.fs.unit.liq_outlet.pressure[0]))
        assert (pytest.approx(7549.4, abs=1e-1) ==
                value(iapws.fs.unit.liq_outlet.enth_mol[0]))
        assert (pytest.approx(59.532, abs=1e-3) ==
                value(iapws.fs.unit.liq_outlet.flow_mol[0]))

    @pytest.mark.initialize
    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    def test_conservation(self, iapws):
        assert abs(value(iapws.fs.unit.inlet.flow_mol[0] -
                         (iapws.fs.unit.vap_outlet.flow_mol[0] +
                          iapws.fs.unit.liq_outlet.flow_mol[0]))) <= 1e-6
        assert abs(value(iapws.fs.unit.inlet.flow_mol[0] *
                         iapws.fs.unit.inlet.enth_mol[0] -
                         (iapws.fs.unit.vap_outlet.flow_mol[0] *
                          iapws.fs.unit.vap_outlet.enth_mol[0] +
                          iapws.fs.unit.liq_outlet.flow_mol[0] *
                          iapws.fs.unit.liq_outlet.enth_mol[0]))) <= 1e-6

    @pytest.mark.ui
    def test_report(self, iapws):
        iapws.fs.unit.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Steam turbine outlet stage model.  This model is based on:

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136.
"""
__Author__ = "John Eslick"

from pyomo.common.config import In
from pyomo.environ import (Var, Expression, Constraint, sqrt, SolverFactory,
                           value, Param)
from pyomo.opt import TerminationCondition

from idaes.core import declare_process_block_class
from idaes.unit_models.pressure_changer import (PressureChangerData,
                                                ThermodynamicAssumption)
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

_log = getIdaesLogger(__name__)


@declare_process_block_class("TurbineOutletStage",
    doc="Outlet stage steam turbine model")
class TurbineOutletStageData(PressureChangerData):
    # Same settings as the default pressure changer, but force to expander with
    # isentropic efficiency
    CONFIG = PressureChangerData.CONFIG()
    CONFIG.compressor = False
    CONFIG.get('compressor')._default = False
    CONFIG.get('compressor')._domain = In([False])
    CONFIG.thermodynamic_assumption = ThermodynamicAssumption.isentropic
    CONFIG.get('thermodynamic_assumption')._default = \
        ThermodynamicAssumption.isentropic
    CONFIG.get('thermodynamic_assumption')._domain = \
        In([ThermodynamicAssumption.isentropic])
    def build(self):
        super(TurbineOutletStageData, self).build()

        self.flow_coeff = Var(initialize=0.0333,
            doc="Turbine flow coefficient [kg*C^0.5/s/Pa]")
        self.eff_dry = Var(initialize=0.87,
            doc="Turbine dry isentropic efficiency")
        self.design_exhaust_flow_vol = Var(initialize=6000.0,
            doc="Design exit volumetirc flowrate [m^3/s]")
        self.efficiency_mech = Var(initialize=0.98,
            doc="Turbine mechanical efficiency")
        self.flow_scale = Param(mutable=True, default=1e-4, doc=
            "Scaling factor for pressure flow relation should be approximatly"
            " the same order of magnitude as the expected flow.")
        self.eff_dry.fix()
        self.design_exhaust_flow_vol.fix()
        self.flow_coeff.fix()
        self.efficiency_mech.fix()
        self.ratioP[:] = 1 # make sure these have a number value
        self.deltaP[:] = 0 #   to avoid an error later in initialize

        @self.Expression(self.flowsheet().config.time,
                         doc="Efficiency factor correlation")
        def tel(b, t):
            f = b.control_volume.properties_out[t].flow_vol/b.design_exhaust_flow_vol
            return 1e6*(-0.0035*f**5 + 0.022*f**4 - 0.0542*f**3 + 0.0638*f**2 -
                        0.0328*f + 0.0064)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Equation: Stodola, for choked flow")
        def stodola_equation(b, t):
            flow = b.control_volume.properties_in[t].flow_mol
            mw = b.control_volume.properties_in[t].mw
            Tin = b.control_volume.properties_in[t].temperature
            Pin = b.control_volume.properties_in[t].pressure
            Pr = b.ratioP[t]
            cf = b.flow_coeff
            return (b.flow_scale**2)*flow**2*mw**2*(Tin - 273.15) == \
                (b.flow_scale**2)*cf**2*Pin**2*(1 - Pr**2)

        @self.Expression(
            self.flowsheet().config.time,
            doc="Equation: isentropic specific enthalpy change"
        )
        def delta_enth_isentropic(b, t):
            flow = b.control_volume.properties_in[t].flow_mol
            work_isen = b.work_isentropic[t]
            return work_isen/flow

        @self.Constraint(self.flowsheet().config.time,
                         doc="Equation: Efficiency correlation")
        def efficiency_correlation(b, t):
            x = b.control_volume.properties_out[t].vapor_frac
            eff = b.efficiency_isentropic[t]
            dh_isen = b.delta_enth_isentropic[t]
            tel = b.tel[t]
            return eff == b.eff_dry*x*(1 - 0.65*(1 - x))*(1 + tel/dh_isen)

        @self.Expression(self.flowsheet().config.time,
                         doc="Thermodynamic power [J/s]")
        def power_thermo(b, t):
            return b.control_volume.work[t]

        @self.Expression(self.flowsheet().config.time,
                         doc="Shaft power [J/s]")
        def power_shaft(b, t):
            return b.power_thermo[t]*b.efficiency_mech

    def _get_performance_contents(self, time_point=0):
        pc = super()._get_performance_contents(time_point=time_point)
        pc["vars"]["Mechanical Efficiency"] = self.efficiency_mech
        pc["vars"]["Flow Coefficient"] = self.flow_coeff
        pc["vars"]["Isentropic Efficieincy (Dry)"] = self.eff_dry
        pc["vars"]["Design Exhaust Flow"] = self.design_exhaust_flow_vol

        pc["exprs"] = {}
        pc["exprs"]["Thermodynamic Power"] = self.power_thermo[time_point]
        pc["exprs"]["Shaft Power"] = self.power_shaft[time_point]

        pc["params"] = {}
        pc["params"]["Flow Scaling"] = self.flow_scale

        return pc

    def initialize(self, state_args={}, outlvl=6, solver='ipopt',
        optarg={'tol': 1e-6, 'max_iter':30}):
        """
        Initialize the outlet turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves.

        Args:
            state_args (dict): Initial state for property initialization
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        init_log = getInitLogger(self.name, outlvl)
        # sp is what to save to make sure state after init is same as the start
        #   saves value, fixed, and active state, doesn't load originally free
        #   values, this makes sure original problem spec is same but initializes
        #   the values of free vars
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)
        # Deactivate special constraints
        self.stodola_equation.deactivate()
        self.efficiency_correlation.deactivate()
        self.deltaP.unfix()
        self.ratioP.unfix()
        # Fix turbine parameters + eff_isen
        self.eff_dry.fix()
        self.design_exhaust_flow_vol.fix()
        self.flow_coeff.fix()

        # fix inlet and free outlet
        for t in self.flowsheet().config.time:
            for k, v in self.inlet.vars.items():
                v[t].fix()
            for k, v in self.outlet.vars.items():
                v[t].unfix()
            # If there isn't a good guess for efficiency or outlet pressure
            # provide something reasonable.
            eff = self.efficiency_isentropic[t]
            eff.fix(eff.value if value(eff) > 0.3 and value(eff) < 1.0 else 0.8)
            # for outlet pressure try outlet pressure, pressure ratio, delta P,
            # then if none of those look reasonable use a pressure ratio of 0.8
            # to calculate outlet pressure
            Pout = self.outlet.pressure[t]
            Pin = self.inlet.pressure[t]
            prdp = value((self.deltaP[t] - Pin)/Pin)
            if value(Pout/Pin) > 0.95 or value(Pout/Pin) < 0.003:
                if value(self.ratioP[t]) < 0.9 and value(self.ratioP[t]) > 0.01:
                    Pout.fix(value(Pin*self.ratioP))
                elif prdp < 0.9 and prdp > 0.01:
                    Pout.fix(value(prdp*Pin))
                else:
                    Pout.fix(value(Pin*0.3))
            else:
                Pout.fix()
        self.deltaP[:] = value(Pout - Pin)
        self.ratioP[:] = value(Pout/Pin)

        for t in self.flowsheet().config.time:
            self.properties_isentropic[t].pressure.value = \
                value(self.outlet.pressure[t])
            self.properties_isentropic[t].flow_mol.value = \
                value(self.inlet.flow_mol[t])
            self.properties_isentropic[t].enth_mol.value = \
                value(self.inlet.enth_mol[t]*0.95)
            self.outlet.flow_mol[t].value = \
                value(self.inlet.flow_mol[t])
            self.outlet.enth_mol[t].value = \
                value(self.inlet.enth_mol[t]*0.95)

        # Make sure the initialization problem has no degrees of freedom
        # This shouldn't happen here unless there is a bug in this
        dof = degrees_of_freedom(self)
        try:
            assert(dof == 0)
        except:
            init_log.log(5, "Error: Degrees_of_freedom = {}".format(dof))
            raise

        mw = self.control_volume.properties_in[0].mw
        Tin = self.control_volume.properties_in[0].temperature
        Pin = self.control_volume.properties_in[0].pressure
        Pr = self.ratioP[0]
        cf = self.flow_coeff
        self.inlet.flow_mol.fix(value(cf*Pin*sqrt(1 - Pr**2)/mw/sqrt(Tin - 273.15)))

        # one bad thing about reusing this is that the log messages aren't
        # really compatible with being nested inside another initialization
        super().initialize(
            state_args=state_args,
            outlvl=outlvl,
            solver=solver,
            optarg=optarg,
        )
        # Free eff_isen and activate sepcial constarints
        self.efficiency_isentropic.unfix()
        self.outlet.pressure.fix()
        self.inlet.flow_mol.unfix()
        self.stodola_equation.activate()
        self.efficiency_correlation.activate()

        slvr = SolverFactory(solver)
        slvr.options = optarg
        res = slvr.solve(self, tee=init_tee(init_log))
        init_log.log(5, "Initialization Complete (Outlet Stage): {}".format(condition(res)))

        # reload original spec
        from_json(self, sd=istate, wts=sp)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test surrogate modeling.
"""
# stdlib
import logging
import sys

# third-party
import pytest

# local
from idaes.dmf import surrmod, errors
from idaes.dmf.experiment import Experiment

# for testing
from .util import init_logging, tmp_dmf

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


@pytest.fixture
def model_data():
    return {"scoops": [1, 2, 3, 4], "cost": [0.5, 0.9, 1.2, 1.4]}


# Real ALAMO


@pytest.mark.skipif(not surrmod.alamo_enabled, reason="ALAMO is disabled")
def test_init(tmp_dmf):
    _test_init(tmp_dmf)


def _test_init(tmp_dmf):
    _ = surrmod.SurrogateModel(Experiment(tmp_dmf))


@pytest.mark.skipif(not surrmod.alamo_enabled, reason="ALAMO is disabled")
def test_run(tmp_dmf, model_data):
    _test_run(tmp_dmf, model_data)


def _test_run(tmp_dmf, model_data):
    sml = surrmod.SurrogateModel(Experiment(tmp_dmf))
    sml.set_input_data(model_data, ["scoops"], "cost")
    try:
        _ = sml.run()
    except errors.AlamoError:
        raise


# Mock ALAMO


class AlamoMock(object):
    def __init__(self):
        self.doalamo_called = False

    def doalamo(self, x, z, *args, **kwargs):
        self.doalamo_called = True
        self.xdata = x
        self.zdata = z
        return {}

    def __enter__(self):
        """Monkey-patch 'surrmod' to use this class
        """
        surrmod.alamo_enabled = True
        surrmod.alamopy = self  # redirect calls to alamopy.doalamo()
        return self

    def __exit__(self, *args):
        surrmod.alamo_enabled = False
        surrmod.alamopy = None


@pytest.mark.skipif(surrmod.alamo_enabled, reason="ALAMO works, no mocking")
def test_init_mock(tmp_dmf):
    with AlamoMock() as mock:
        _test_init(tmp_dmf)
        assert not mock.doalamo_called


@pytest.mark.skipif(surrmod.alamo_enabled, reason="ALAMO works, no mocking")
def test_run_mock(tmp_dmf, model_data):
    with AlamoMock() as mock:
        _test_run(tmp_dmf, model_data)
        assert mock.doalamo_called
        assert tuple(mock.xdata) == tuple(model_data["scoops"])
        assert tuple(mock.zdata) == tuple(model_data["cost"])


@pytest.mark.skipif(surrmod.alamo_enabled, reason="ALAMO works, no mocking")
def test_bad_columns_mock(tmp_dmf, model_data):
    with AlamoMock() as mock:
        m = surrmod.SurrogateModel(Experiment(tmp_dmf))
        with pytest.raises(KeyError):
            m.set_input_data(model_data, ["snork"], "cost")



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for DMF Jupyter "magics"
"""
# stdlib
import os
import sys
import webbrowser

# third-party
import pytest

# local
from idaes.dmf import magics, DMF
from idaes.dmf.magics import DMFMagicError
from idaes.dmf.resource import Resource
from .util import TempDir

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

# monkey-patch webbrowser to do nothing
webbrowser.open_new = lambda url: None


class MockShell(object):
    """Mock object for IPython 'shell'.
    """

    def ev(self, name):
        """Mock evaluation.
        For most names, return the DMF class.
        For the special name "FAIL", raise an exception.
        For the special name "NODOCS", return a non-documented Python module.
        """
        if name == "FAIL":
            raise ValueError("Failure")
        elif name == "NODOCS":
            return os
        return DMF


###############################################################################
# Fixtures


@pytest.fixture
def magics_impl():
    shell = MockShell()
    return magics.DmfMagicsImpl(shell)


@pytest.fixture
def tmp_dmf():
    with TempDir() as d:
        dmf = DMF(d, name="tempdmf", create=True)
        yield dmf
        del dmf


@pytest.fixture
def tmp_magics():
    with TempDir() as d:
        magics_impl.dmf_init(d, "create")
        yield magics_impl
        del magics_impl


###############################################################################
# Tests


def test_init_create(magics_impl):
    with TempDir() as d:
        magics_impl.dmf_init(d, "create")
        assert magics_impl.initialized


def test_init_existing(tmp_dmf, magics_impl):
    magics_impl.dmf_init(tmp_dmf.root)


def test_init_extraignored(tmp_dmf, magics_impl):
    magics_impl.dmf_init(tmp_dmf.root, "foo")  # just generate warning
    magics_impl.dmf_init(tmp_dmf.root, "foo", "bar")  # ditto


def test_init_required(magics_impl):
    pytest.raises(DMFMagicError, magics_impl.dmf_info)
    pytest.raises(DMFMagicError, magics_impl.dmf_help, "anything")


def test_init_badpath(magics_impl):
    nosuchpath = os.path.join(os.path.sep, *map(str, range(10)))
    pytest.raises(DMFMagicError, magics_impl.dmf_init, nosuchpath)
    pytest.raises(DMFMagicError, magics_impl.dmf_init, nosuchpath, "create")


def test_init_goodpath(magics_impl):
    with TempDir() as goodpath:
        magics_impl.dmf_init(goodpath, "create")


def test_dmf_cmd(magics_impl):
    pytest.raises(DMFMagicError, magics_impl.dmf, "not a command")
    pytest.raises(DMFMagicError, magics_impl.dmf, "list stuff")
    magics_impl.dmf("")  # Empty value is OK


def test_dmf_workspaces(magics_impl):
    with TempDir() as goodpath:
        # Try with a good path, but no workspace
        # so should return "No valid workspaces found"
        msg = magics_impl.dmf_workspaces(goodpath)
        # make sure first token of response is not a number
        token = msg.split()[0]
        pytest.raises(ValueError, int, token)
        # Create a workspace
        magics_impl.dmf_init(goodpath, "create")
        # Try again, should return "1 workspace found" (or similar)
        msg = magics_impl.dmf_workspaces(goodpath)
        token = msg.split()[0]
        assert int(token) > 0


def test_dmf_list(magics_impl):
    # Start before setting up DMF; should not work
    pytest.raises(DMFMagicError, magics_impl.dmf, "list")
    # Now set up the DMF, add 1 resource, call again. This time,
    # should be OK
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        rsrc = Resource()
        magics_impl._dmf.add(rsrc)  # XXX: fragile since it accesses _dmf
        magics_impl.dmf("list")


def test_dmf_info_initrequired(magics_impl):
    # should fail with no DMF
    pytest.raises(DMFMagicError, magics_impl.dmf_info)


def test_dmf_info_topics(magics_impl):
    # should fail with DMF, and topics (not implemented)
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        pytest.raises(DMFMagicError, magics_impl.dmf_info, "uptime")


def test_dmf_info_notopics(magics_impl):
    # should succeed with DMF, and no topics
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        magics_impl.dmf_info()


def test_dmf_info_extrameta(magics_impl):
    # by filling in the metadata, this will test the dmf_info code
    # that prints out list and dict data structures
    with TempDir() as wspath:
        open(os.path.join(wspath, DMF.WORKSPACE_CONFIG), "w").write(
            """
_id: this-is-a-temporary-config
logging:
    idaes.dmf.dmfbase:
        level: debug
        output: _stderr_
    root:
        output: _stdout_
    dmf:
        output: _stdout_
    .dmf.experiment:
        output: _stdout_
    # equivalent to previous
    idaes.dmf.experiment:
        output: /tmp/experiment.log
    # user
    crazy.little.logger:
        level: error
        output: _stderr_
        """
        )
        magics_impl.dmf_init(wspath)
        magics_impl.dmf_info()


def test_dmf_help(magics_impl):
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        magics_impl.dmf_help()


def test_dmf_help_badargs(magics_impl):
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        # This will generate a warning, but is OK. Only the first
        # object is tried (which fails; another warning)
        result = magics_impl.dmf_help("this", "that")
        assert result is None  # None means "OK"


def test_dmf_help_obj(magics_impl):
    with TempDir() as wspath:
        magics_impl.dmf_init(wspath, "create")
        # "special" names
        for name in "dmf", "help", "idaes":
            magics_impl.dmf_help(name)
        # object
        magics_impl.dmf_help("idaes.dmf.dmfbase.DMF")
        # failure (still returns None)
        assert magics_impl.dmf_help("FAIL") is None




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# This subroutine generates target rates of genration for each species
# Currently, alamo is called explicitly in RIPE
# alamopy dependencies exist through calls to alamopy.multos in order
# to manipulate files alamo creates
# soubroutines in this file
# doalamo         - generate alamo models for dynamic targets
# dopwalamo       - wrapper for pw alamo application
# gentartgs       - generate target values from process data
# sstargets       - called in gentargets for steady-state
# dynamictargets  - called in gentargets for dynamic data

import numpy as np
import os


def doalamo(data, time, xlo, xup, rpc, rspec, sharedata):
    # This subroutine calls alamo to generate targets for dynamic problems
    # this call could be replaced by a call to alamopy.doalamo()
    # Inputs:
    # data     - process data from ripemodel()
    # time     - time from the pc dictionary in ripemodel()
    # xlo/xup  - bounds for time in dynamic problems
    # rpc      - list containing range of process conditions
    # rspec    - list of species
    # Outputs:
    # profiles - dictionary of prfiles from recursive piece-wise alamo
    # r2ret    - r2 of submodels

    # import sympy
    from idaes.surrogate.alamopy import multos as mos
    from idaes.surrogate.ripe import debug
    # from sympy import symbols

    if len(np.shape(data)) == 2:
        data = np.expand_dims(data, axis=-1)

    # Convert data to python index
    xlo = xlo - 1
    xup = xup - 1

    r2ret = {}
    profiles = {}
    # st = symbols("st")
    # Build alamo models for each process condition
    for p in rpc:
        r2ret[p] = {}
        profiles[p] = {}
        with open("temp.alm", "w") as a:
            a.write("linfcns 1\n")
            a.write("funform 1\n")
            a.write("trace 1\n")
            a.write("tracefname temptrace.trc\n")
            a.write("solvemip 1\n")
            a.write("ninputs 1\n")
            a.write("noutputs " + str(len(rspec)) + "\n")
            a.write("xlabels st\n")
            a.write("modeler 1\n")
            a.write("monomialpower 0.5 1 1.5 2 \n")
            a.write("logfcns 1\n")
            a.write("expfcns 1\n")
            a.write("constant 0\n")
            a.write("xmin 0\n")
            a.write("xmax " + str(time[-1]) + "\n")
            a.write("cvxbic 1\n")
            a.write("initialpoints " + str(len(range(xlo, xup))) + "\n")
            a.write("ndata " + str(len(range(xlo, xup))) + "\n")
            a.write("begin_data\n")
            for i in range(xlo, xup):  #
                tstr = str(time[i]) + " "
                for j in rspec:
                    tstr = tstr + str(data[i, j, p]) + " "
                a.write(tstr + "\n")
            a.write("end_data\n")
        os.system(sharedata["alamo_path"] + " temp.alm > tmpscratch")

    f = open("temptrace.trc")
    lf = f.read()
    f.close()
    lf2 = lf.split("\n")
    z = 0
    for i in rpc:
        for j in rspec:
            if "#filename" in lf2[z]:
                z = z + 1
            profiles[i][j] = lf2[z].split(",")[-1].split(" = ")[-1]
            r2ret[i][j] = lf2[z].split(",")[18]
            z = z + 1
    mos.copyfile("sv.alm", "temp.alm")
    if not debug["savescratch"]:
        mos.deletefile("sv.alm", "tmpscratch", "temp.alm", "temptrace.trc", "temp.lst")

    return [profiles, r2ret, sharedata]


def dopwalamo(data, tdata, profile, r2, xlo, xup, rpc, rspec, sharedata):
    from sympy import symbols
    from sympy.parsing.sympy_parser import parse_expr

    # debug = ripe.debug

    # python syntax of data indicies
    xlm = xlo - 1
    xum = xup - 1

    st = symbols("st")
    for pc in rpc:
        for s in rspec:
            amodel = parse_expr(profile[pc][s])
            if float(r2[pc][s]) > sharedata["pwalamo"]["err"]:
                sharedata["pwalamo"]["profiles"][pc][s].append(profile[pc][s])
                sharedata["pwalamo"]["tlo"][pc][s].append(xlo)
                sharedata["pwalamo"]["tup"][pc][s].append(xup)
                sharedata["pwalamo"]["etrack"][pc][s].append(r2)
            elif (xup - xlo) > 2 * sharedata["pwalamo"]["nmin"]:
                cutoff = (1.0 - float(sharedata["pwalamo"]["err"])) * np.sum(
                    (data[xlm:xum, s, pc] - np.average(data[xlm:xum, s, pc])) ** 2
                )
                temp = 0
                i = xlm
                while temp < 0.5 * cutoff:
                    temp = temp + (data[i, s, pc] - amodel.subs(st, tdata[i])) ** 2
                    i = i + 1
                if (i - xlo) < sharedata["pwalamo"]["nmin"]:
                    i = xlo + sharedata["pwalamo"]["nmin"]
                if (xup - i) > sharedata["pwalamo"]["nmin"]:
                    i = xup - sharedata["pwalamo"]["nmin"]
                [lp1, lr1, sharedata] = doalamo(
                    data, tdata, xlo, i, [pc], [s], sharedata
                )
                sharedata = dopwalamo(
                    data, tdata, lp1, lr1, xlo, i, [pc], [s], sharedata
                )
                [lp2, lr2, sharedata] = doalamo(
                    data, tdata, i, xup, [pc], [s], sharedata
                )
                sharedata = dopwalamo(
                    data, tdata, lp2, lr2, i, xup, [pc], [s], sharedata
                )
            else:
                sharedata["pwalamo"]["profiles"][pc][s].append(profile[pc][s])
                sharedata["pwalamo"]["tlo"][pc][s].append(xlo)
                sharedata["pwalamo"]["tup"][pc][s].append(xup)
                sharedata["pwalamo"]["etrack"][pc][s].append(r2[pc][s])
    return sharedata


def gentargets(data, kwargs, fdata, pc, sd):
    # this subroutine is a wrapper for gene
    # Determine target rate values
    # Sharedata is passed to track pw-alamo progress
    if "t" in kwargs.keys():
        targets, sd = dynamictargets(data, fdata, pc, sd)
    else:
        targets = sstargets(fdata, pc)

    if sd["zscale"]:
        s_targets = np.max([np.abs(np.ndarray.flatten(targets))])
        savetargets = targets
        targets = np.divide(targets, s_targets)
    else:
        s_targets = 1.0
        savetargets = targets

    return targets, s_targets, savetargets, sd


def sstargets(fdata, pc):
    # Determine steady-state target values via mass balance
    # 0 = (flow/volume) * (Conc_out - Conc_in)
    n, s = np.shape(fdata)
    targets = np.zeros([n, s])
    for i in range(n):
        for j in range(s):
            targets[i, j] = (pc["flow"][i][j] / pc["vol"][i]) * (
                fdata[i, j] - pc["x0"][i][j]
            )
    return targets


def dynamictargets(data, fdata, pc, sharedata):
    # import sympy
    from sympy import symbols, diff

    n, spec = np.shape(fdata)
    dshape = np.shape(data)
    ntimedata = dshape[0]
    targets = np.ones([n, spec])
    tempn = n
    st = symbols("st")

    while tempn >= ntimedata:
        [prof, r2, sharedata] = doalamo(
            data[tempn - ntimedata : tempn, :, :],
            pc["t"][tempn - ntimedata : tempn],
            1,
            ntimedata,
            range(int(pc["npc"])),
            range(int(spec)),
            sharedata,
        )
        for l in list(["profiles", "tlo", "tup", "etrack"]):
            for prc in range(pc["npc"]):
                sharedata["pwalamo"][l][prc] = {}
                for s in range(spec):
                    sharedata["pwalamo"][l][prc][s] = list()
        sharedata = dopwalamo(
            data[tempn - ntimedata : tempn, :, :],
            pc["t"][tempn - ntimedata : tempn],
            prof,
            r2,
            1,
            ntimedata,
            range(pc["npc"]),
            range(spec),
            sharedata,
        )

        for p in range(spec):
            for i in range(pc["npc"]):
                pwt = 0
                for j in range(tempn - ntimedata, tempn):
                    profile = sharedata["pwalamo"]["profiles"][i][p][pwt]
                    targets[j, p] = diff(profile, st).subs(st, pc["t"][j])

                    if j == sharedata["pwalamo"]["tup"][i][p][pwt]:
                        pwt = pwt + 1

        tempn -= ntimedata
    return targets, sharedata



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
General purpose mixer block for IDAES models
"""
from enum import Enum

from pyomo.environ import (Constraint,
                           Param,
                           PositiveReals,
                           Reals,
                           RangeSet,
                           SolverFactory,
                           TerminationCondition,
                           Var)
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core import (declare_process_block_class,
                        UnitModelBlockData,
                        useDefault,
                        MaterialBalanceType)
from idaes.core.util.config import (is_physical_parameter_block,
                                    is_state_block,
                                    list_of_strings)
from idaes.core.util.exceptions import (BurntToast,
                                        ConfigurationError,
                                        PropertyNotSupportedError)
from idaes.core.util.math import smooth_min
from idaes.core.util.misc import add_object_reference
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.logger import getIdaesLogger, getInitLogger, init_tee, condition

__author__ = "Andrew Lee"


# Set up logger
_log = getIdaesLogger(__name__)


# Enumerate options for balances
class MixingType(Enum):
    none = 0
    extensive = 1


class MomentumMixingType(Enum):
    none = 0
    minimize = 1
    equality = 2
    minimize_and_equality = 3


@declare_process_block_class("Mixer")
class MixerData(UnitModelBlockData):
    """
    This is a general purpose model for a Mixer block with the IDAES modeling
    framework. This block can be used either as a stand-alone Mixer unit
    operation, or as a sub-model within another unit operation.

    This model creates a number of StateBlocks to represent the incoming
    streams, then writes a set of phase-component material balances, an
    overall enthalpy balance and a momentum balance (2 options) linked to a
    mixed-state StateBlock. The mixed-state StateBlock can either be specified
    by the user (allowing use as a sub-model), or created by the Mixer.

    When being used as a sub-model, Mixer should only be used when a set
    of new StateBlocks are required for the streams to be mixed. It should not
    be used to mix streams from mutiple ControlVolumes in a single unit model -
    in these cases the unit model developer should write their own mixing
    equations.
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Mixer blocks are always steady-state."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Mixer blocks do not contain holdup, thus this must be
False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for mixer",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))
    CONFIG.declare("inlet_list", ConfigValue(
        domain=list_of_strings,
        description="List of inlet names",
        doc="""A list containing names of inlets,
**default** - None.
**Valid values:** {
**None** - use num_inlets argument,
**list** - a list of names to use for inlets.}"""))
    CONFIG.declare("num_inlets", ConfigValue(
        domain=int,
        description="Number of inlets to unit",
        doc="""Argument indicating number (int) of inlets to construct, not
used if inlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use inlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of inlets to create (will be named with sequential integers
from 1 to num_inlets).}"""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("has_phase_equilibrium", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Calculate phase equilibrium in mixed stream",
        doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}"""))
    CONFIG.declare("energy_mixing_type", ConfigValue(
        default=MixingType.extensive,
        domain=MixingType,
        description="Method to use when mixing energy flows",
        doc="""Argument indicating what method to use when mixing energy
flows of incoming streams,
**default** - MixingType.extensive.
**Valid values:** {
**MixingType.none** - do not include energy mixing equations,
**MixingType.extensive** - mix total enthalpy flows of each phase.}"""))
    CONFIG.declare("momentum_mixing_type", ConfigValue(
        default=MomentumMixingType.minimize,
        domain=MomentumMixingType,
        description="Method to use when mixing momentum/pressure",
        doc="""Argument indicating what method to use when mixing momentum/
pressure of incoming streams,
**default** - MomentumMixingType.minimize.
**Valid values:** {
**MomentumMixingType.none** - do not include momentum mixing equations,
**MomentumMixingType.minimize** - mixed stream has pressure equal to the
minimimum pressure of the incoming streams (uses smoothMin operator),
**MomentumMixingType.equality** - enforces equality of pressure in mixed and
all incoming streams.,
**MomentumMixingType.minimize_and_equality** - add constraints for pressure
equal to the minimum pressure of the inlets and constraints for equality of
pressure in mixed and all incoming streams. When the model is initially built,
the equality constraints are deactivated.  This option is useful for switching
between flow and pressure driven simulations.}"""))
    CONFIG.declare("mixed_state_block", ConfigValue(
        default=None,
        domain=is_state_block,
        description="Existing StateBlock to use as mixed stream",
        doc="""An existing state block to use as the outlet stream from the
Mixer block,
**default** - None.
**Valid values:** {
**None** - create a new StateBlock for the mixed stream,
**StateBlock** - a StateBock to use as the destination for the mixed stream.}
"""))
    CONFIG.declare("construct_ports", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Construct inlet and outlet Port objects",
        doc="""Argument indicating whether model should construct Port objects
linked to all inlet states and the mixed state,
**default** - True.
**Valid values:** {
**True** - construct Ports for all states,
**False** - do not construct Ports."""))

    def build(self):
        """
        General build method for MixerData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(MixerData, self).build()

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        # Create list of inlet names
        inlet_list = self.create_inlet_list()

        # Build StateBlocks
        inlet_blocks = self.add_inlet_state_blocks(inlet_list)

        if self.config.mixed_state_block is None:
            mixed_block = self.add_mixed_state_block()
        else:
            mixed_block = self.get_mixed_state_block()

        mb_type = self.config.material_balance_type
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = \
                mixed_block[t_ref].default_material_balance_type()

        if mb_type != MaterialBalanceType.none:
            self.add_material_mixing_equations(inlet_blocks=inlet_blocks,
                                               mixed_block=mixed_block,
                                               mb_type=mb_type)
        else:
            raise BurntToast("{} received unrecognised value for "
                             "material_mixing_type argument. This "
                             "should not occur, so please contact "
                             "the IDAES developers with this bug."
                             .format(self.name))

        if self.config.energy_mixing_type == MixingType.extensive:
            self.add_energy_mixing_equations(inlet_blocks=inlet_blocks,
                                             mixed_block=mixed_block)
        elif self.config.energy_mixing_type == MixingType.none:
            pass
        else:
            raise ConfigurationError("{} received unrecognised value for "
                                     "material_mixing_type argument. This "
                                     "should not occur, so please contact "
                                     "the IDAES developers with this bug."
                                     .format(self.name))

        if self.config.momentum_mixing_type == MomentumMixingType.minimize:
            self.add_pressure_minimization_equations(inlet_blocks=inlet_blocks,
                                                     mixed_block=mixed_block)
        elif self.config.momentum_mixing_type == MomentumMixingType.equality:
            self.add_pressure_equality_equations(inlet_blocks=inlet_blocks,
                                                 mixed_block=mixed_block)
        elif self.config.momentum_mixing_type == \
                MomentumMixingType.minimize_and_equality:
            self.add_pressure_minimization_equations(inlet_blocks=inlet_blocks,
                                                     mixed_block=mixed_block)
            self.add_pressure_equality_equations(inlet_blocks=inlet_blocks,
                                                 mixed_block=mixed_block)
            self.pressure_equality_constraints.deactivate()
        elif self.config.momentum_mixing_type == MomentumMixingType.none:
            pass
        else:
            raise ConfigurationError("{} recieved unrecognised value for "
                                     "momentum_mixing_type argument. This "
                                     "should not occur, so please contact "
                                     "the IDAES developers with this bug."
                                     .format(self.name))

        self.add_port_objects(inlet_list, inlet_blocks, mixed_block)

    def create_inlet_list(self):
        """
        Create list of inlet stream names based on config arguments.

        Returns:
            list of strings
        """
        if (self.config.inlet_list is not None and
                self.config.num_inlets is not None):
            # If both arguments provided and not consistent, raise Exception
            if len(self.config.inlet_list) != self.config.num_inlets:
                raise ConfigurationError(
                        "{} Mixer provided with both inlet_list and "
                        "num_inlets arguments, which were not consistent ("
                        "length of inlet_list was not equal to num_inlets). "
                        "PLease check your arguments for consistency, and "
                        "note that it is only necessary to provide one of "
                        "these arguments.".format(self.name))
        elif self.config.inlet_list is None and self.config.num_inlets is None:
            # If no arguments provided for inlets, default to num_inlets = 2
            self.config.num_inlets = 2

        # Create a list of names for inlet StateBlocks
        if self.config.inlet_list is not None:
            inlet_list = self.config.inlet_list
        else:
            inlet_list = ['inlet_' + str(n)
                          for n in range(1, self.config.num_inlets+1)]

        return inlet_list

    def add_inlet_state_blocks(self, inlet_list):
        """
        Construct StateBlocks for all inlet streams.

        Args:
            list of strings to use as StateBlock names

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True

        # Create empty list to hold StateBlocks for return
        inlet_blocks = []

        # Create an instance of StateBlock for all inlets
        for i in inlet_list:
            i_obj = self.config.property_package.state_block_class(
                        self.flowsheet().config.time,
                        doc="Material properties at inlet",
                        default=tmp_dict)

            setattr(self, i+"_state", i_obj)

            inlet_blocks.append(getattr(self, i+"_state"))

        return inlet_blocks

    def add_mixed_state_block(self):
        """
        Constructs StateBlock to represent mixed stream.

        Returns:
            New StateBlock object
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = \
            self.config.has_phase_equilibrium
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False

        self.mixed_state = self.config.property_package.state_block_class(
                                self.flowsheet().config.time,
                                doc="Material properties of mixed stream",
                                default=tmp_dict)

        return self.mixed_state

    def get_mixed_state_block(self):
        """
        Validates StateBlock provided in user arguments for mixed stream.

        Returns:
            The user-provided StateBlock or an Exception
        """
        # Sanity check to make sure method is not called when arg missing
        if self.config.mixed_state_block is None:
            raise BurntToast("{} get_mixed_state_block method called when "
                             "mixed_state_block argument is None. This should "
                             "not happen.".format(self.name))

        # Check that the user-provided StateBlock uses the same prop pack
        if (self.config.mixed_state_block[
                self.flowsheet().config.time.first()].config.parameters
                != self.config.property_package):
            raise ConfigurationError(
                    "{} StateBlock provided in mixed_state_block argument "
                    " does not come from the same property package as "
                    "provided in the property_package argument. All "
                    "StateBlocks within a Mixer must use the same "
                    "property package.".format(self.name))

        return self.config.mixed_state_block

    def add_material_mixing_equations(self,
                                      inlet_blocks,
                                      mixed_block,
                                      mb_type):
        """
        Add material mixing equations.
        """
        # Get phase component list(s)
        phase_component_list = self._get_phase_comp_list()

        if mb_type == MaterialBalanceType.componentPhase:
            # Create equilibrium generation term and constraints if required
            if self.config.has_phase_equilibrium is True:
                # Get units from property package
                units = {}
                for u in ['holdup', 'time']:
                    try:
                        units[u] = (self.config.property_package
                                    .get_metadata().default_units[u])
                    except KeyError:
                        units[u] = '-'

                try:
                    self.phase_equilibrium_generation = Var(
                            self.flowsheet().config.time,
                            self.config.property_package.phase_equilibrium_idx,
                            domain=Reals,
                            doc="Amount of generation in unit by phase "
                                "equilibria [{}/{}]"
                                .format(units['holdup'], units['time']))
                except AttributeError:
                    raise PropertyNotSupportedError(
                        "{} Property package does not contain a list of phase "
                        "equilibrium reactions (phase_equilibrium_idx), "
                        "thus does not support phase equilibrium."
                        .format(self.name))

            # Define terms to use in mixing equation
            def phase_equilibrium_term(b, t, p, j):
                if self.config.has_phase_equilibrium:
                    sd = {}
                    for r in b.config.property_package.phase_equilibrium_idx:
                        if b.config.property_package.\
                                phase_equilibrium_list[r][0] == j:
                            if b.config.property_package.\
                                    phase_equilibrium_list[r][1][0] == p:
                                sd[r] = 1
                            elif b.config.property_package.\
                                    phase_equilibrium_list[r][1][1] == p:
                                sd[r] = -1
                            else:
                                sd[r] = 0
                        else:
                            sd[r] = 0

                    return sum(b.phase_equilibrium_generation[t, r]*sd[r]
                               for r in
                               b.config.property_package.phase_equilibrium_idx)
                else:
                    return 0

            # Write phase-component balances
            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.phase_list,
                             self.config.property_package.component_list,
                             doc="Material mixing equations")
            def material_mixing_equations(b, t, p, j):
                if j in phase_component_list[p]:
                    return 0 == (
                        sum(inlet_blocks[i][t].get_material_flow_terms(p, j)
                            for i in range(len(inlet_blocks))) -
                        mixed_block[t].get_material_flow_terms(p, j) +
                        phase_equilibrium_term(b, t, p, j))
                else:
                    return Constraint.Skip

        elif mb_type == MaterialBalanceType.componentTotal:
            # Write phase-component balances
            @self.Constraint(self.flowsheet().config.time,
                             self.config.property_package.component_list,
                             doc="Material mixing equations")
            def material_mixing_equations(b, t, j):
                return 0 == sum(
                    sum(inlet_blocks[i][t].get_material_flow_terms(p, j)
                        for i in range(len(inlet_blocks))) -
                    mixed_block[t].get_material_flow_terms(p, j)
                    for p in b.config.property_package.phase_list)

        elif mb_type == MaterialBalanceType.total:
            # Write phase-component balances
            @self.Constraint(self.flowsheet().config.time,
                             doc="Material mixing equations")
            def material_mixing_equations(b, t):
                return 0 == sum(sum(
                    sum(inlet_blocks[i][t].get_material_flow_terms(p, j)
                        for i in range(len(inlet_blocks))) -
                    mixed_block[t].get_material_flow_terms(p, j)
                    for j in b.config.property_package.component_list)
                    for p in b.config.property_package.phase_list)
        elif mb_type == MaterialBalanceType.elementTotal:
            raise ConfigurationError("{} Mixers do not support elemental "
                                     "material balances.".format(self.name))
        elif mb_type == MaterialBalanceType.none:
            pass
        else:
            raise BurntToast("{} Mixer received unrecognised value for "
                             "material_balance_type. This should not happen, "
                             "please report this bug to the IDAES developers."
                             .format(self.name))

    def add_energy_mixing_equations(self, inlet_blocks, mixed_block):
        """
        Add energy mixing equations (total enthalpy balance).
        """
        self.scaling_factor_energy = Param(
                        default=1e-6,
                        mutable=True,
                        doc='Energy balance scaling parameter')

        @self.Constraint(self.flowsheet().config.time, doc="Energy balances")
        def enthalpy_mixing_equations(b, t):
            return 0 == self.scaling_factor_energy*(
                sum(sum(inlet_blocks[i][t].get_enthalpy_flow_terms(p)
                    for p in b.config.property_package.phase_list)
                    for i in range(len(inlet_blocks))) -
                sum(mixed_block[t].get_enthalpy_flow_terms(p)
                    for p in b.config.property_package.phase_list))

    def add_pressure_minimization_equations(self, inlet_blocks, mixed_block):
        """
        Add pressure minimization equations. This is done by sequential
        comparisons of each inlet to the minimum pressure so far, using
        the IDAES smooth minimum fuction.
        """
        if not hasattr(self, "inlet_idx"):
            self.inlet_idx = RangeSet(len(inlet_blocks))
        # Add variables
        self.minimum_pressure = Var(self.flowsheet().config.time,
                                    self.inlet_idx,
                                    doc='Variable for calculating '
                                        'minimum inlet pressure')

        self.eps_pressure = Param(mutable=True,
                                  initialize=1e-3,
                                  domain=PositiveReals,
                                  doc='Smoothing term for '
                                      'minimum inlet pressure')

        # Calculate minimum inlet pressure
        @self.Constraint(self.flowsheet().config.time,
                         self.inlet_idx,
                         doc='Calculation for minimum inlet pressure')
        def minimum_pressure_constraint(b, t, i):
            if i == self.inlet_idx.first():
                return self.minimum_pressure[t, i] == (
                           inlet_blocks[i-1][t].pressure)
            else:
                return self.minimum_pressure[t, i] == (
                        smooth_min(self.minimum_pressure[t, i-1],
                                   inlet_blocks[i-1][t].pressure,
                                   self.eps_pressure))

        # Set inlet pressure to minimum pressure
        @self.Constraint(self.flowsheet().config.time,
                         doc='Link pressure to control volume')
        def mixture_pressure(b, t):
            return mixed_block[t].pressure == (
                       self.minimum_pressure[t,
                                             self.inlet_idx.last()])

    def add_pressure_equality_equations(self, inlet_blocks, mixed_block):
        """
        Add pressure equality equations. Note that this writes a number of
        constraints equal to the number of inlets, enforcing equality between
        all inlets and the mixed stream.
        """
        if not hasattr(self, "inlet_idx"):
            self.inlet_idx = RangeSet(len(inlet_blocks))

        # Create equality constraints
        @self.Constraint(self.flowsheet().config.time,
                         self.inlet_idx,
                         doc='Calculation for minimum inlet pressure')
        def pressure_equality_constraints(b, t, i):
            return mixed_block[t].pressure == inlet_blocks[i-1][t].pressure

    def add_port_objects(self, inlet_list, inlet_blocks, mixed_block):
        """
        Adds Port objects if required.

        Args:
            a list of inlet StateBlock objects
            a mixed state StateBlock object

        Returns:
            None
        """
        if self.config.construct_ports is True:
            # Add ports
            for p in inlet_list:
                i_state = getattr(self, p+"_state")
                self.add_port(name=p, block=i_state, doc="Inlet Port")
            self.add_port(name="outlet", block=mixed_block, doc="Outlet Port")

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().config.time:
            try:
                inlet_list = blk.create_inlet_list()
                for i in inlet_list:
                    i_block = getattr(blk, i+"_state")
                    i_block[t].model_check()
            except AttributeError:
                _log.warning('{} Mixer inlet property block has no model '
                             'checks. To correct this, add a model_check '
                             'method to the associated StateBlock class.'
                             .format(blk.name))
            try:
                if blk.config.mixed_state_block is None:
                    blk.mixed_state[t].model_check()
                else:
                    blk.config.mixed_state_block.model_check()
            except AttributeError:
                _log.warning('{} Mixer outlet property block has no '
                             'model checks. To correct this, add a '
                             'model_check method to the associated '
                             'StateBlock class.'.format(blk.name))

    def use_minimum_inlet_pressure_constraint(self):
        """Activate the mixer pressure = mimimum inlet pressure constraint and
        deactivate the mixer pressure and all inlet pressures are equal
        constraints. This should only be used when momentum_mixing_type ==
        MomentumMixingType.minimize_and_equality.
        """
        if self.config.momentum_mixing_type != \
                MomentumMixingType.minimize_and_equality:
            _log.warning(
                """use_minimum_inlet_pressure_constraint() can only be used
                when momentum_mixing_type ==
                MomentumMixingType.minimize_and_equality""")
            return
        self.minimum_pressure_constraint.activate()
        self.pressure_equality_constraints.deactivate()

    def use_equal_pressure_constraint(self):
        """Deactivate the mixer pressure = mimimum inlet pressure constraint
        and activate the mixer pressure and all inlet pressures are equal
        constraints. This should only be used when momentum_mixing_type ==
        MomentumMixingType.minimize_and_equality.
        """
        if self.config.momentum_mixing_type != \
                MomentumMixingType.minimize_and_equality:
            _log.warning(
                """use_equal_pressure_constraint() can only be used when
                momentum_mixing_type ==
                MomentumMixingType.minimize_and_equality""")
            return
        self.minimum_pressure_constraint.deactivate()
        self.pressure_equality_constraints.activate()

    def initialize(blk, outlvl=6, optarg={},
                   solver='ipopt', hold_state=False):
        '''
        Initialization routine for mixer (default solver ipopt)

        Keyword Arguments:
            outlvl : sets output level of initialization routine
                 * 0 = Use default idaes.init logger setting
                 * 1 = Maximum output
                 * 2 = Include solver output
                 * 3 = Return solver state for each step in subroutines
                 * 4 = Return solver state for each step in routine
                 * 5 = Final initialization status and exceptions
                 * 6 = No output
            optarg : solver options dictionary object (default={})
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        '''
        init_log = getInitLogger(blk.name, outlvl)
        # Set solver options
        opt = SolverFactory(solver)
        opt.options = optarg

        # Initialize inlet state blocks
        flags = {}
        inlet_list = blk.create_inlet_list()
        i_block_list = []
        for i in inlet_list:
            i_block = getattr(blk, i+"_state")
            i_block_list.append(i_block)
            flags[i] = {}
            flags[i] = i_block.initialize(outlvl=outlvl+1,
                                          optarg=optarg,
                                          solver=solver,
                                          hold_state=True)

        # Initialize mixed state block
        if blk.config.mixed_state_block is None:
            mblock = blk.mixed_state
        else:
            mblock = blk.config.mixed_state_block


        o_flags = {}
        # Calculate initial guesses for mixed stream state
        for t in blk.flowsheet().config.time:
            # Iterate over state vars as defined by property package
            s_vars = mblock[t].define_state_vars()
            for s in s_vars:
                i_vars = []
                for k in s_vars[s]:
                    # Record whether variable was fixed or not
                    o_flags[t, s, k] = s_vars[s][k].fixed

                    # If fixed, use current value
                    # otherwise calculate guess from mixed state
                    if not s_vars[s][k].fixed:
                        for i in range(len(i_block_list)):
                            i_vars.append(getattr(i_block_list[i][t],
                                                  s_vars[s].local_name))

                        if s == "pressure":
                            # If pressure, use minimum as initial guess
                            mblock[t].pressure.value = min(
                                    i_block_list[i][t].pressure.value
                                    for i in range(len(i_block_list)))
                        elif "flow" in s:
                            # If a "flow" variable (i.e. extensive), sum inlets
                            for k in s_vars[s]:
                                s_vars[s][k].value = sum(
                                        i_vars[i][k].value
                                        for i in range(len(i_block_list)))
                        else:
                            # Otherwise use average of inlets
                            for k in s_vars[s]:
                                s_vars[s][k].value = (
                                    sum(i_vars[i][k].value for i in
                                        range(len(i_block_list))) /
                                    len(i_block_list))

        mblock.initialize(outlvl=outlvl+1,
                          optarg=optarg,
                          solver=solver,
                          hold_state=False)

        # Revert fixed status of variables to what they were before
        for t in blk.flowsheet().config.time:
            for s in s_vars:
                for k in s_vars[s]:
                    s_vars[s][k].fixed = o_flags[t, s, k]

        if blk.config.mixed_state_block is None:
            if (hasattr(blk, "pressure_equality_constraints") and
                    blk.pressure_equality_constraints.active is True):
                blk.pressure_equality_constraints.deactivate()
                for t in blk.flowsheet().config.time:
                    sys_press = getattr(
                            blk,
                            blk.create_inlet_list()[0]+"_state")[t].pressure
                    blk.mixed_state[t].pressure.fix(sys_press.value)

                results = opt.solve(blk, tee=init_tee(init_log))

                blk.pressure_equality_constraints.activate()
                for t in blk.flowsheet().config.time:
                    blk.mixed_state[t].pressure.unfix()

            else:
                results = opt.solve(blk, tee=init_tee(init_log))

            init_log.log(5, "Initialization Complete: {}".format(condition(results)))
        else:
            init_log.log(5, "Initialization Complete.")

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl+1)

    def release_state(blk, flags, outlvl=6):
        '''
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        '''
        inlet_list = blk.create_inlet_list()
        for i in inlet_list:
            i_block = getattr(blk, i+"_state")
            i_block.release_state(flags[i], outlvl=outlvl+1)

    def _get_stream_table_contents(self, time_point=0):
        io_dict = {}
        inlet_list = self.create_inlet_list()
        for i in inlet_list:
            io_dict[i] = getattr(self, i+"_state")
        if self.config.mixed_state_block is None:
            io_dict["Outlet"] = self.mixed_state
        else:
            io_dict["Outlet"] = self.config.mixed_state_block
        return create_stream_table_dataframe(
                io_dict,
                time_point=time_point)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Property data types.

Ability to import, etc. from text files is
part of the methods in the type.

Import property database from textfile(s):
* See :meth:`PropertyData.from_csv`, for the expected format for data.
* See :meth:`PropertyMetadata()` for the expected format for metadata.

"""
# stdlib
import csv
import json
import logging

# third-party
try:
    import pandas as pd
    import numpy as np
except ImportError:
    np, pd = None, None

# local
from .util import get_file
from . import tabular

__author__ = 'Dan Gunter'

_log = logging.getLogger(__name__)


class AddedCSVColumnError(KeyError):
    """Error for :meth:PropertyData.add_csv()
    """

    def __init__(self, names, how_bad, column_type=''):
        ctype = column_type + ' ' if column_type else ''
        if len(names) == 1:
            msg = 'Added CSV data {} {}column "{}"'.format(
                how_bad, ctype, list(names)[0]
            )
        else:
            msg = 'Added CSV data {} {}columns: {}'.format(
                how_bad, ctype, ', '.join(list(names))
            )
        KeyError.__init__(self, msg)


class Fields(tabular.Fields):
    """Constants for fields.
    """

    # Values for "type" field
    C_STATE, C_PROP = 'state', 'property'


class PropertyTable(tabular.Table):
    """Property data and metadata together (at last!)
    """

    def __init__(self, data=None, **kwargs):
        """Constructor.
        """
        if isinstance(data, PropertyData):
            pdata = data
        elif isinstance(data, list):
            pdata = PropertyData(data)
        else:
            raise TypeError('list or PropertyData object required')
        super(PropertyTable, self).__init__(data=pdata, **kwargs)

    @classmethod
    def load(cls, file_or_path, validate=True):
        """Create PropertyTable from JSON input.

        Args:
            file_or_path (file or str): Filename or file object
                from which to read the JSON-formatted data.
            validate (bool): If true, apply validation to input JSON data.

        Example input::

            {
                "meta": [
                    {"datatype": "MEA",
                     "info": "J. Chem. Eng. Data, 2009, Vol 54, pg. 306-310",
                     "notes": "r is MEA weight fraction in aqueous soln.",
                     "authors": "Amundsen, T.G., Lars, E.O., Eimer, D.A.",
                     "title": "Density and Viscosity of ..."}
                ],
                "data": [
                    {"name": "Viscosity Value",
                     "units": "mPa-s",
                     "values": [2.6, 6.2],
                     "error_type": "absolute",
                     "errors": [0.06, 0.004],
                     "type": "property"},
                    {"name": "r",
                     "units": "",
                     "values": [0.2, 1000],
                     "type": "state"}
                ]
            }
        """
        fp = get_file(file_or_path)
        d = json.load(fp)
        PropertyTable._validate_json(d)
        metalist = d[Fields.META]
        meta = [PropertyMetadata(m) for m in metalist]
        data = PropertyData(d[Fields.DATA])
        tbl = PropertyTable(data=data)
        for m in meta:
            tbl.add_metadata(m)
        return tbl


class PropertyData(tabular.TabularData):
    """Class representing property data that knows how to
    construct itself from a CSV file.

    You can build objects from multiple CSV files as well.
    See the property database section of the API docs for
    details, or read the code in :meth:`add_csv` and the
    tests in :mod:`idaes_dmf.propdb.tests.test_mergecsv`.
    """

    embedded_units = r'(.*)\((.*)\)'

    def __init__(self, data):
        """Construct new object from input list.

        Example input::

            [{
                "name": "Density Data",
                "units": "g/cm^3",
                "values": [1.0053, 1.0188, .., ],
                "errors": [.00005, .., .00005],
                "error_type": "absolute",
                "type": "property"
            }, ...etc...]

        Args:
            data (list): Input data columns

        Returns:
            (PropertyData) New instance.
        """
        super(PropertyData, self).__init__(data, error_column=True)
        self._nstates = len(self.states)

    @property
    def states(self):
        return [c for c in self.columns if self._is_state(c)]

    @property
    def properties(self):
        return [c for c in self.columns if self._is_prop(c)]

    @staticmethod
    def _is_state(c):
        return c[Fields.COLTYPE] == Fields.C_STATE

    @staticmethod
    def _is_prop(c):
        return c[Fields.COLTYPE] == Fields.C_PROP

    def names(self, states=True, properties=True):
        """Get column names.

        Args:
            states (bool): If False, exclude "state" data, e.g. the
                          ambient temperature, and only
                          include measured property values.
            properties (bool): If False, excluse property data

        Returns:
            list[str]: List of column names.
        """
        result = []
        if states:
            result.extend([v[Fields.DATA_NAME] for v in self.states])
        if properties:
            result.extend([v[Fields.DATA_NAME] for v in self.properties])
        return result

    def is_state_column(self, index):
        """Whether given column is state.

        Args:
            index (int): Index of column
        Returns:
            (bool) State or property and the column number.
        Raises:
            IndexError: No column at that index.
        """
        col = self.columns[index]
        return self._is_state(col)

    def is_property_column(self, index):
        """Whether given column is a property. See :meth:`is_state_column`."""
        return not self.is_state_column(index)

    def as_arr(self, states=True):
        """Export property data as arrays.

        Args:
            states (bool): If False, exclude "state" data, e.g. the
                          ambient temperature, and only
                          include measured property values.
        Returns:
            (values[M,N], errors[M,N]) Two arrays of floats,
            each with M columns having N values.
        Raises:
            ValueError if the columns are not all the same length
        """
        n, values, errors = None, [], []
        # extract state columns
        if states:
            for v in self.states:
                vals = v[Fields.DATA_VALUES]
                if n is None:
                    n = len(vals)
                elif len(vals) != n:
                    raise ValueError(
                        'State values "{}" length {} != {}'.format(
                            v[Fields.DATA_NAME], len(vals), n
                        )
                    )
                values.append(vals)
                errors.append([0] * len(vals))
        # extract property columns
        for v in self.properties:
            vals = v[Fields.DATA_VALUES]
            if n is None:
                n = len(vals)
            elif len(vals) != n:
                raise ValueError(
                    'Property values "{}" length {} != {}'.format(
                        v[Fields.DATA_NAME], len(vals), n
                    )
                )
            values.append(v[Fields.DATA_VALUES])
            errors.append(v[Fields.DATA_ERRORS])
        return values, errors

    def values_dataframe(self, states=True):
        """Get values as a dataframe.

        Args:
            states (bool): see :meth:`names()`.

        Returns:
            (pd.DataFrame) Pandas dataframe for values.
 
        Raises:
            ImportError: If `pandas` or `numpy` were never
                successfully imported.
        """
        return self._get_prop_dataframe(Fields.DATA_VALUES, states)

    def errors_dataframe(self, states=False):
        """Get errors as a dataframe.

        Args:
            states (bool): If False, exclude state data.
                           This is the default, because states do not
                           normally have associated error information.
        Returns:
            pd.DataFrame: Pandas dataframe for values.

        Raises:
            ImportError: If `pandas` or `numpy` were never
                successfully imported.
        """
        return self._get_prop_dataframe(Fields.DATA_ERRORS, states)

    def _get_prop_dataframe(self, field, states):
        self._check_pandas_import()
        a1, names = [], []
        if states:
            a1 = [v[field] for v in self.states]
            names = [v[Fields.DATA_NAME] for v in self.states]
        a1.extend([v[field] for v in self.properties])
        names.extend([v[Fields.DATA_NAME] for v in self.properties])
        a2 = np.array(a1).transpose()
        return pd.DataFrame(a2, columns=names)

    @staticmethod
    def from_csv(file_or_path, nstates=0):
        """Import the CSV data.

        Expected format of the  files is a header plus data rows.

        Header: Index-column,  Column-name(1), Error-column(1),  \
                Column-name(2), Error-column(2), ..
        Data: <index>, <val>, <errval>, <val>, <errval>, ..

        Column-name is in the format "Name (units)"

        Error-column is in the format "<type> Error", where "<type>" is
        the error type.

        Args:
            file_or_path (file-like or str): Input file
            nstates (int): Number of state columns, appearing
               first before property columns.

        Returns:
            PropertyData: New properties instance
        """
        input_file = get_file(file_or_path)
        csv_file = csv.reader(input_file)
        row = next(csv_file)
        names, data = PropertyData._prop_parse_csv_headers(nstates, row)
        for row in csv_file:
            # print('@@ parse csv row: {}'.format(row))
            PropertyData._parse_csv_row(data, row, error_column=True)
        obj = PropertyData(data)
        return obj

    def add_csv(self, file_or_path, strict=False):
        """Add to existing object from a new CSV file.

        Depending on the value of the `strict` argument (see
        below), the new file may or may not have the same 
        properties as the object -- but it always needs to have
        the same number of state columns, and in the same order.

        .. note:: Data that is "missing" because of property columns in
           one CSV and not the other will be filled with `float(nan)` values.
        
        Args:
            file_or_path (file or str): Input file. This should be in exactly
                the same format as expected by :meth:from_csv().
            strict (bool): If true, require that the columns in the input
                CSV match columns in this object. Otherwise, only require
                that *state* columns in input CSV match columns in this
                object. New property columns are added, and matches
                to existing property columns will append the data.
                
            
        Raises:
            AddedCSVColumnError: If the new CSV column headers are not the
                same as the ones in this object.
            
        Returns:
            (int) Number of added rows
        """
        nstates = self._nstates
        input_file = get_file(file_or_path)
        csv_file = csv.reader(input_file)

        # Parse the header
        row = next(csv_file)
        hdr_names, hdr_data = PropertyData._prop_parse_csv_headers(nstates, row)

        # print('@@ add_csv, column names = {}, data columns = {}'
        #      .format(hdr_names, self.names()))

        # Check that set of keys in new data is the same
        cur_keys = set(self.names())
        new_keys = set(hdr_names)
        # This is used to re-order input data
        rowmap = None
        if strict:
            if cur_keys > new_keys:
                missing = cur_keys - new_keys
                raise AddedCSVColumnError(missing, 'is missing')
            elif new_keys > cur_keys:
                extra = new_keys - cur_keys
                raise AddedCSVColumnError(extra, 'has extra')
            elif new_keys != cur_keys:
                extra = new_keys - cur_keys
                missing = cur_keys - new_keys
                namelist = (
                    '(' + ','.join(extra) + ')',
                    'instead of',
                    '(' + ','.join(missing) + ')',
                )
                raise AddedCSVColumnError(namelist, 'has different')
        else:
            # check that all states are in common
            hdr_states = filter(self._is_state, hdr_data)
            new_states = [s[Fields.DATA_NAME] for s in hdr_states]
            new_states = set(new_states)
            cur_states = set(self.names(properties=False))
            if new_states != cur_states:
                extra = new_states - cur_states
                missing = cur_states - new_states
                if extra and missing:
                    namelist = (
                        '(' + ','.join(extra) + ')',
                        'instead of',
                        '(' + ','.join(missing) + ')',
                    )
                    raise AddedCSVColumnError(
                        namelist, 'has different', column_type='state'
                    )
                elif extra:
                    raise AddedCSVColumnError(extra, 'has extra', column_type='state')
                elif missing:
                    raise AddedCSVColumnError(
                        missing, 'is missing', column_type='state'
                    )
                else:
                    raise RuntimeError('unexpected branch')
            # check that at least one property is in common
            new_prop = new_keys - new_states
            if not new_prop:
                return 0  # no data
            cur_prop = set(self.names(states=False))

            # Add columns for all properties only found on the input,
            # and initialize values to a list of NaN's as long as the
            # current table, so data in all fields will be the same length.

            # Initialize rowmap with mapping for state columns
            rowmap = [-1] * len(hdr_names)
            idx = 0
            for i, s in enumerate(hdr_data):
                if s[Fields.COLTYPE] == Fields.C_PROP:
                    continue
                rowmap[i] = idx
                idx += 1

            nan_list = [float('nan')] * self.num_rows
            idx = 0
            for i, value in enumerate(hdr_data):
                if value[Fields.COLTYPE] == Fields.C_STATE:
                    continue
                name = value[Fields.DATA_NAME]
                if name not in cur_prop:
                    value[Fields.DATA_NAME] = name
                    value[Fields.DATA_VALUES] = nan_list[:]
                    value[Fields.DATA_ERRORS] = nan_list[:]
                    value[Fields.COLTYPE] = Fields.C_PROP
                    self._data.append(value)
                    rowmap[i] = len(self.properties) - 1
                else:
                    rowmap[i] = idx + self._nstates
                idx += 1

        # print("@@ rowmap = {}".format(rowmap))

        # Parse the new data
        num_added = 0
        new_rowlen = 1 + 2 * len(self.names())
        for row in csv_file:
            if rowmap:
                # Re-order according to the rowmap.
                # By initializing with NaN, any columns not in the
                # input, but in the current data, will be replaced with NaN
                # values.
                row2 = [float('nan')] * new_rowlen
                # print('@@ row={} row2-init={}'.format(row, row2))
                for i, j in enumerate(rowmap):
                    row2[j * 2 + 1] = row[i * 2 + 1]  # value
                    row2[j * 2 + 2] = row[i * 2 + 2]  # error
                row = row2
            self._parse_csv_row(self._data, row, error_column=True)
            num_added += 1
            self._nrows += 1

        return num_added

    @classmethod
    def _prop_parse_csv_headers(cls, nstates, headers):
        """Parse a row of CSV headers which are pairs
        of columns like "<name> [(units)], <error-type> Error".

        Returns:
             (names, data). Names is a list of all the column names.
             Data is a dict with two keys, "properties" and "states".
             Each value will be a list of property/state objects.
        """
        names, data = cls._parse_csv_headers(headers, error_column=True)
        for i in range(0, nstates):
            data[i][Fields.COLTYPE] = Fields.C_STATE
        for i in range(nstates, len(data)):
            data[i][Fields.COLTYPE] = Fields.C_PROP
        return names, data


class PropertyMetadata(tabular.Metadata):
    """Class to import property metadata.
    """

    pass


class PropertyColumn(tabular.Column):
    """Data column for a property.
    """

    type_name = 'Property'

    def __init__(self, name, data):
        tabular.Column.__init__(self, name, data)
        self.errors = data[Fields.DATA_ERRORS]
        self.error_type = data[Fields.DATA_ERRTYPE]

    def data(self):
        return {
            Fields.DATA_UNITS: self.units,
            Fields.DATA_VALUES: self.values,
            Fields.DATA_ERRORS: self.errors,
            Fields.DATA_ERRTYPE: self.error_type,
        }


class StateColumn(tabular.Column):
    """Data column for a state.
    """

    type_name = 'State'

    def __init__(self, name, data):
        tabular.Column.__init__(self, name, data)
        self.errors = [0.0] * len(self)
        self.error_type = 'none'

    def data(self):
        return {Fields.DATA_UNITS: self.units, Fields.DATA_VALUES: self.values}


def convert_csv(meta_csv, datatype, data_csv, nstates, output):
    meta = PropertyMetadata.from_csv(meta_csv)
    meta.datatype = datatype
    data = PropertyData.from_csv(data_csv, nstates)
    obj = PropertyTable(data=data, metadata=meta)
    ofile = get_file(output, mode='w')
    obj.dump(ofile)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for idaes.dmf.experiment module
"""
import logging
import pytest
import sys

from idaes.dmf import experiment, errors, resource
from .util import tmp_dmf  # noqa -- this is a fixture
from .util import init_logging, tmp_dmf

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

if sys.platform.startswith("win"):
    pytest.skip("skipping DMF tests on Windows", allow_module_level=True)

init_logging()
_log = logging.getLogger(__name__)


def test_init(tmp_dmf):
    exp = experiment.Experiment(tmp_dmf, name="try1", desc="Nice try")
    assert exp.v["name"] == "try1"
    assert exp.id


def test_create_many(tmp_dmf):
    for i in range(100):
        exp = experiment.Experiment(tmp_dmf, name="try{i}".format(i=i))
        assert exp.v["name"] == "try{i}".format(i=i)
        assert exp.id


def test_remove_workflow(tmp_dmf):
    # A workflow of copy/remove
    # make an experiment
    e1 = experiment.Experiment(tmp_dmf, name="one", version="0.0.1")
    # make a new version of it
    e2 = e1.copy(version="0.0.2")
    # link the two together
    e1.link(e2, predicate=resource.PR_VERSION)
    # remove the first one (what happens to the link?)
    e1.remove()
    # check that the first one can't be used any more
    with pytest.raises(errors.BadResourceError):
        e1.update()
    with pytest.raises(errors.BadResourceError):
        e1.link(e2, predicate=resource.PR_VERSION)
    # check that the copy can still be modified
    e2.v["desc"] = "This is a copy of e1"
    e2.update()  # this fixes relations in the DB
    # check that the link is gone, i.e. there are no
    # relations in e2 any more
    assert len(e2.v["relations"]) == 0




##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests ProcessBlock and ProcessBlockData.

Author: John Eslick
"""
import pytest
from pyomo.environ import ConcreteModel, Var, value
from pyomo.common.config import ConfigValue
from idaes.core import (
    ProcessBlock,
    ProcessBlockData,
    declare_process_block_class)


@declare_process_block_class("MyBlock")
class MyBlockData(ProcessBlockData):
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare("xinit", ConfigValue(default=1001, domain=float))
    CONFIG.declare("yinit", ConfigValue(default=1002, domain=float))
    def build(self):
        super(MyBlockData, self).build()
        self.x = Var(initialize=self.config.xinit)
        self.y = Var(initialize=self.config.yinit)


class TestProcessBlock(object):
    def test_scalar_noargs(self):
        m = ConcreteModel()
        m.b = MyBlock()
        assert(isinstance(m.b.x, Var))
        assert(isinstance(m.b.y, Var))
        assert(value(m.b.x) == 1001)
        assert(value(m.b.y) == 1002)

    def test_vec_noargs(self):
        m = ConcreteModel()
        m.b = MyBlock([1,2,3])
        assert(isinstance(m.b[1].x, Var))
        assert(isinstance(m.b[1].y, Var))
        assert(isinstance(m.b[2].x, Var))
        assert(isinstance(m.b[2].y, Var))
        assert(isinstance(m.b[3].x, Var))
        assert(isinstance(m.b[3].y, Var))
        assert(value(m.b[1].x) == 1001)
        assert(value(m.b[1].y) == 1002)
        assert(value(m.b[2].x) == 1001)
        assert(value(m.b[2].y) == 1002)
        assert(value(m.b[3].x) == 1001)
        assert(value(m.b[3].y) == 1002)

    def test_scalar_args1(self):
        m = ConcreteModel()
        m.b = MyBlock(default={"xinit":1, "yinit":2})
        assert(isinstance(m.b.x, Var))
        assert(isinstance(m.b.y, Var))
        assert(value(m.b.x) == 1)
        assert(value(m.b.y) == 2)

    def test_scalar_args2(self):
        m = ConcreteModel()
        m.b = MyBlock(initialize={None:{"xinit":1, "yinit":2}})
        assert(isinstance(m.b.x, Var))
        assert(isinstance(m.b.y, Var))
        assert(value(m.b.x) == 1)
        assert(value(m.b.y) == 2)

    def test_vec_args(self):
        m = ConcreteModel()
        m.b = MyBlock(
            [1,2,3],
            default={"xinit":1, "yinit":2},
            initialize={2:{"xinit":2001, "yinit":2002}})
        assert(isinstance(m.b[1].x, Var))
        assert(isinstance(m.b[1].y, Var))
        assert(isinstance(m.b[2].x, Var))
        assert(isinstance(m.b[2].y, Var))
        assert(isinstance(m.b[3].x, Var))
        assert(isinstance(m.b[3].y, Var))
        assert(m.b[1].index() == 1)
        assert(value(m.b[1].x) == 1)
        assert(value(m.b[1].y) == 2)
        assert(value(m.b[2].x) == 2001)
        assert(value(m.b[2].y) == 2002)
        assert(value(m.b[3].x) == 1)
        assert(value(m.b[3].y) == 2)

    def test_user_map(self):
        m = ConcreteModel()
        def new_imap(i):
            if i > 0 and i < 4:
                return 1
            else:
                return i

        m.b = MyBlock(
            [0,1,2,3,4],
            idx_map = new_imap,
            initialize={0:{"xinit":2001, "yinit":2002},
                        1:{"xinit":5001, "yinit":5002},
                        2:{"xinit":6001, "yinit":6002},
                        4:{"xinit":7001, "yinit":7002}})
        assert(value(m.b[0].x) == 2001)
        assert(value(m.b[0].y) == 2002)
        assert(value(m.b[1].x) == 5001)
        assert(value(m.b[1].y) == 5002)
        assert(value(m.b[2].x) == 5001) # although this index (2) is in initialize
        assert(value(m.b[2].y) == 5002) # the idx_map function maps it to 1
        assert(value(m.b[3].x) == 5001)
        assert(value(m.b[3].y) == 5002)
        assert(value(m.b[4].x) == 7001)
        assert(value(m.b[4].y) == 7002)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Standard IDAES flash model.
"""
# Import Python libraries
import logging
from pandas import DataFrame

# Import Pyomo libraries
from pyomo.environ import Constraint, value
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Port

# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault)
from idaes.unit_models.separator import (Separator,
                                         SplittingType,
                                         EnergySplittingType)

from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.misc import add_object_reference

__author__ = "Andrew Lee, Jaffer Ghouse"


# Set up logger
logger = logging.getLogger('idaes.unit_model')


@declare_process_block_class("Flash")
class FlashData(UnitModelBlockData):
    """
    Standard Flash Unit Model Class
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Flash units do not support dynamic behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Flash units do not have defined volume, thus
this must be False."""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("energy_split_basis", ConfigValue(
        default=EnergySplittingType.equal_temperature,
        domain=EnergySplittingType,
        description="Type of constraint to write for energy splitting",
        doc="""Argument indicating basis to use for splitting energy this is
not used for when ideal_separation == True.
**default** - EnergySplittingType.equal_temperature.
**Valid values:** {
**EnergySplittingType.equal_temperature** - outlet temperatures equal inlet
**EnergySplittingType.equal_molar_enthalpy** - oulet molar enthalpies equal
inlet,
**EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
flows.}"""))
    CONFIG.declare("ideal_separation", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Ideal splitting flag",
        doc="""Argument indicating whether ideal splitting should be used.
Ideal splitting assumes perfect separation of material, and attempts to
avoid duplication of StateBlocks by directly partitioning outlet flows to
ports,
**default** - True.
**Valid values:** {
**True** - use ideal splitting methods. Cannot be combined with
has_phase_equilibrium = True,
**False** - use explicit splitting equations with split fractions.}"""))
    CONFIG.declare("has_heat_transfer", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Heat transfer term construction flag",
        doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
constructed,
**default** - True.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(FlashData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(default={
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.control_volume.add_state_blocks(
            has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=True)

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer)

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Add Ports
        self.add_inlet_port()

        split_map = {}
        for p in self.config.property_package.phase_list:
            split_map[p] = p

        self.split = Separator(default={
                "property_package": self.config.property_package,
                "property_package_args": self.config.property_package_args,
                "outlet_list": ["Vap", "Liq"],
                "split_basis": SplittingType.phaseFlow,
                "ideal_separation": self.config.ideal_separation,
                "ideal_split_map": split_map,
                "mixed_state_block": self.control_volume.properties_out,
                "has_phase_equilibrium": not self.config.ideal_separation,
                "energy_split_basis": self.config.energy_split_basis})
        if not self.config.ideal_separation:
            def split_frac_rule(b, t, o):
                return b.split.split_fraction[t, o, o] == 1
            self.split_fraction_eq = Constraint(self.flowsheet().config.time,
                                                self.split.outlet_idx,
                                                rule=split_frac_rule)

        self.vap_outlet = Port(extends=self.split.Vap)
        self.liq_outlet = Port(extends=self.split.Liq)

        # Add references
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != 'none'):
            add_object_reference(self, "heat_duty", self.control_volume.heat)
        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != 'none'):
            add_object_reference(self, "deltaP", self.control_volume.deltaP)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}

        for n, v in {"Inlet": "inlet",
                     "Vapor Outlet": "vap_outlet",
                     "Liquid Outlet": "liq_outlet"}.items():
            port_obj = getattr(self, v)

            stream_attributes[n] = {}

            for k in port_obj.vars:
                for i in port_obj.vars[k].keys():
                    if isinstance(i, float):
                        stream_attributes[n][k] = value(
                                port_obj.vars[k][time_point])
                    else:
                        if len(i) == 2:
                            kname = str(i[1])
                        else:
                            kname = str(i[1:])
                        stream_attributes[n][k+" "+kname] = \
                            value(port_obj.vars[k][time_point, i[1:]])

        return DataFrame.from_dict(stream_attributes, orient="columns")
